# Copyright 2014 Adobe. All Rights Reserved.

"""
How to build a feature-file format.
1) Traverse script->language->feature tree, building up list as we go.

These are all expected to be in alphabetic order in the font.

Add each  feature entry as FeatDict[tag][lang_sys].append(index]), and as
FeatDict[index] .append(lang_sys)

If there is no DFLT script, then step through all entries of  FeatDict[index].
Add FeatDict[tag]["DFLT_dflt"].append(index) for all FeatDict[index] where the
number of lang_sys entries is the same as the number of lang_sys values.

Write out the language systems.

Collect and write the class defs.
For each lookup in lookup list: for each class def in the lookup:
    add entry to ClassesByNameDict[<glyph list as tuple>].append(lookup index,
        subtable_index, class index)

Get list of keys for ClassesByNameDict. Sort. MakeClassNameDict.
For each  keys for ClassesByNameDict:
    make class name from <firstglyph_lastglyph>.
    while class name is in ClassesByNameDict:
        find glyph in current class that is not in previous class, and
            make name firstglyph_diff-glyph_last-glyph
    add entry to ClassesByLookup[lookup index, subtable_index, class index)] =
        class name.

Same for mark classes.

Write out class defs.

Write out the feature list.
Make lookupDict = {}
Get feat tag list. For each tag:
    note if DFLT dflt lang sys has any lookups.
    for each lang_sys ( by definition, starts with DFLT, then the rest in
      alpha order).
        If script is DFLT:
            don't write script/lang
            build list of lookups for DFLT script.
        else:
            get union of lookup indicies in all feat indicies
            Remove all the default lookups from the lookup list:
            if no lookups left, continue
            if there was no intersection with default lookups, set exclude
                keyword
            if the script is same as last one, write language only,
            else write script and language.

        if lookupIndex is in lookupDict:
            just write lookup references
        else:
            write lookup definition with name "feat_script_lang_new_index"


"""

__help__ = """
ttxn v1.21.2 Jul 15 2019

Based on the ttx tool, with the same options, except that it is limited to
dumping, and cannot compile. Makes a normalized dump of the font, or of
selected tables.

Adds additional options:
    -nh : suppresses hints in the glyph outline dump.
    -nv : suppress version info in output: head.fontRevision and name ID 3.
    -se : show if lookup is wrapped in an Extension lookup type;
            by default, this information is suppressed.

"""

import sys
from fontTools import ttx
from fontTools.ttLib import TTFont, tagToXML, TTLibError
from fontTools.ttLib.tables.DefaultTable import DefaultTable
from fontTools.misc.loggingTools import Timer
import copy
import subprocess
import re
import collections
import textwrap
import platform
import getopt
import logging

from afdko.fdkutils import get_temp_file_path

__version__ = "1.21.3"

log = logging.getLogger('fontTools.ttx')


curSystem = platform.system()

TX_TOOL = 'tx'


INDENT = "  "
ChainINDENT = INDENT + "C "
MAX_LINE_LENGTH = 80
gtextWrapper = textwrap.TextWrapper(width=MAX_LINE_LENGTH,
                                    subsequent_indent=INDENT + INDENT)

NULL_VALUE_RECORD = "<0 0 0 0>"


def getValueRecord(a, b, c, d):
    return "<%s %s %s %s>" % (a, b, c, d)


def getAnchorString(anchorTable):
    if not anchorTable:
        return "<anchor NULL>"

    tokenList = ["<anchor", "%d" % anchorTable.XCoordinate,
                 "%d" % anchorTable.YCoordinate]
    if anchorTable.Format == 2:
        tokenList.append("%d" % anchorTable.AnchorPoint)
    elif anchorTable.Format == 3:
        tokenList.append("%d" % anchorTable.XDeviceTable)
        tokenList.append("%d" % anchorTable.YDeviceTable)
    return " ".join(tokenList) + ">"


class ClassRecord:

    def __init__(self, lookupIndex, subtableIndex, classIndex, side="",
                 anchor=None):
        self.lookupIndex = lookupIndex
        self.subtableIndex = subtableIndex
        self.classIndex = classIndex
        self.side = side  # None if not a kern pair, else "Left" or "Right"
        self.anchor = anchor

    def __repr__(self):
        return "(lookup %s subtable %s class %s side %s anchor %s)" % (
            self.lookupIndex, self.subtableIndex, self.classIndex, self.side,
            self.anchor)


def addClassDef(otlConv, classDefs, coverage, side=None, anchor=None):
    classDict = {}
    glyphDict = copy.deepcopy(otlConv.glyphDict)

    for name, classIndex in classDefs.items():
        del glyphDict[name]
        try:
            classDict[classIndex].append(name)
        except KeyError:
            classDict[classIndex] = [name]
    classZeroList = None
    if coverage:
        if 0 not in classDict:
            classZeroList = sorted(glyphDict.keys())
            i = len(classZeroList)
            while i > 0:
                i -= 1
                name = classZeroList[i]
                if not (name in coverage.glyphs):
                    classZeroList.remove(name)
    if classZeroList:
        classDict[0] = classZeroList
    else:
        classDict[0] = []

    for classIndex, nameList in classDict.items():
        key = tuple(sorted(nameList))
        classRec = ClassRecord(otlConv.curLookupIndex,
                               otlConv.curSubTableIndex, classIndex, side)
        otlConv.classesByNameList[key].append(classRec)


def AddMarkClassDef(otlConv, markCoverage, markArray, tag):
    markGlyphList = markCoverage.glyphs
    markClass = collections.defaultdict(list)

    for m, markGlyph in enumerate(markGlyphList):
        markRec = markArray.MarkRecord[m]
        markClass[markRec.Class].append(
            (getAnchorString(markRec.MarkAnchor), markGlyph))

    # Get the mark class names from the global dict
    for c, valueList in markClass.items():
        anchorDict = collections.defaultdict(list)
        for anchor, glyph in valueList:
            anchorDict[anchor].append(glyph)
        defList = []
        for anchor, glyphList in anchorDict.items():
            defList.append((tuple(sorted(glyphList)), anchor))
        classRec = ClassRecord(otlConv.curLookupIndex,
                               otlConv.curSubTableIndex, c, tag)
        otlConv.markClassesByDefList[tuple(sorted(defList))].append(classRec)


def classPairGPOS(subtable, otlConv=None):
    addClassDef(otlConv, subtable.ClassDef1.classDefs, subtable.Coverage,
                otlConv.leftSideTag)
    addClassDef(otlConv, subtable.ClassDef2.classDefs, None,
                otlConv.rightSideTag)


def markClassPOS(subtable, otlConv=None):
    AddMarkClassDef(otlConv, subtable.MarkCoverage, subtable.MarkArray,
                    otlConv.markTag)


def markLigClassPOS(subtable, otlConv=None):
    AddMarkClassDef(otlConv, subtable.MarkCoverage, subtable.MarkArray,
                    otlConv.markTag)


def markMarkClassPOS(subtable, otlConv=None):
    AddMarkClassDef(otlConv, subtable.Mark1Coverage, subtable.Mark1Array,
                    otlConv.mark1Tag)


def classContext(subtable, otlConv):
    addClassDef(otlConv, subtable.ClassDef.classDefs, None)


def classChainContext(subtable, otlConv):
    for class_def_name in ('BacktrackClassDef', 'InputClassDef',
                           'LookAheadClassDef'):
        class_def = getattr(subtable, class_def_name)
        if class_def:
            addClassDef(otlConv, class_def.classDefs, None, class_def_name)


def classExt(subtable, otlConv):
    handler = otlConv.classHandlers.get(
        (subtable.ExtensionLookupType, subtable.ExtSubTable.Format), None)
    if handler:
        handler(subtable.ExtSubTable, otlConv)


class ContextRecord:
    def __init__(self, inputList, sequenceIndex):
        self.inputList = inputList
        self.sequenceIndex = sequenceIndex
        self.glyphsUsed = 0
        self.result = ""


def checkGlyphInSequence(glyphName, contextSequence, i):
    retVal = 0
    try:
        input_seq = contextSequence[i]
        if isinstance(input_seq, list):
            if glyphName in input_seq:
                retVal = 1
        elif isinstance(input_seq, str):
            if glyphName == input_seq:
                retVal = 1
    except IndexError:
        pass

    return retVal


def ruleSinglePos(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    if subtable.Format == 1:
        valueString = getValRec(subtable.ValueFormat, subtable.Value)
        if valueString and valueString != "0" and (
                valueString != NULL_VALUE_RECORD):
            for tglyph in subtable.Coverage.glyphs:
                rule = note = None
                if context and (not checkGlyphInSequence(
                        tglyph, inputSeqList, 0)):
                    note = " # Note! Not in input sequence"
                else:
                    note = ""
                rule = "%spos %s %s;%s" % (indent, tglyph, valueString, note)
                if context and note:
                    noMatchRules.append(rule)
                else:
                    rules.append(rule)

    elif subtable.Format == 2:
        for i, subtable_value in enumerate(subtable.Value):
            rule = note = None
            tglyph = subtable.Coverage.glyphs[i]
            valueString = getValRec(subtable.ValueFormat, subtable_value)
            if valueString and valueString != "0" and (
                    valueString != NULL_VALUE_RECORD):
                if context and (tglyph not in inputSeqList):
                    note = " # Note! Not in input sequence"
                else:
                    note = ""
                rule = "%spos %s %s;%s" % (indent, tglyph, valueString, note)
                if context and note:
                    noMatchRules.append(rule)
                else:
                    rules.append(rule)
    else:
        raise AttributeError("Unknown Single Pos format")

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def rulePairPos(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    if subtable.Format == 1:
        firstGlyphList = subtable.Coverage.glyphs
        for i, pairSet in enumerate(subtable.PairSet):
            g1 = firstGlyphList[i]
            for pairValueRec in pairSet.PairValueRecord:
                rule = note = None
                g2 = pairValueRec.SecondGlyph
                note = ""
                if context:
                    note = " # Note! Not in input sequence"
                if pairValueRec.Value1:
                    valueString = getValRec(subtable.ValueFormat1,
                                            pairValueRec.Value1)
                    if valueString and valueString != "0" and (
                            valueString != NULL_VALUE_RECORD):
                        rule = "%spos %s %s %s%s;" % (indent, g1, g2,
                                                      valueString, note)
                if pairValueRec.Value2:
                    valueString = getValRec(subtable.ValueFormat1,
                                            pairValueRec.Value2)
                    if valueString and valueString != "0" and (
                            valueString != NULL_VALUE_RECORD):
                        note = ("%s # Warning: non-zero pos value record for "
                                "second glyph" % note)
                        rule = "%spos %s %s %s;%s" % (indent, g1, g2,
                                                      valueString, note)
                if rule:
                    if context and note:
                        noMatchRules.append(rule)
                    else:
                        rules.append(rule)

    elif subtable.Format == 2:
        for i, classRec1 in enumerate(subtable.Class1Record):
            # if this class reference exists it has to be in classesByLookup.
            g1 = otlConv.classesByLookup[
                otlConv.curLookupIndex, otlConv.curSubTableIndex, i,
                otlConv.leftSideTag]
            for j, classRec2 in enumerate(classRec1.Class2Record):
                rule = note = None
                g2 = otlConv.classesByLookup[
                    otlConv.curLookupIndex, otlConv.curSubTableIndex, j,
                    otlConv.rightSideTag]
                if context:
                    if checkGlyphInSequence(g1, inputSeqList, 0) and (
                            checkGlyphInSequence(g2, inputSeqList, 1)):
                        note = ""
                    else:
                        note = " # Note! Not in input sequence"
                else:
                    note = ""
                if classRec2.Value1:
                    valueString = getValRec(subtable.ValueFormat1,
                                            classRec2.Value1)
                    if valueString and valueString != "0":
                        rule = "%spos %s %s %s;%s" % (indent, g1, g2,
                                                      valueString, note)

                if classRec2.Value2:
                    valueString = getValRec(subtable.ValueFormat1,
                                            classRec2.Value2)
                    if valueString and valueString != "0":
                        note = ("%s # Warning: non-zero pos value record for "
                                "second glyph" % note)
                        rule = "%spos %s %s %s;%s" % (indent, g1, g2,
                                                      valueString, note)
                if rule:
                    if context and note:
                        noMatchRules.append(rule)
                    else:
                        rules.append(rule)
    else:
        raise AttributeError("Unknown Pair Pos format")

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleCursivePos(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    cursiveGlyphList = subtable.Coverage.glyphs
    for i, eRec in enumerate(subtable.EntryExitRecord):
        rule = note = None
        tglyph = cursiveGlyphList[i]
        if context and (not checkGlyphInSequence(tglyph, inputSeqList, 0)):
            note = " # Note! Not in input sequence"
        else:
            note = ""
        anchor1 = getAnchorString(eRec.EntryAnchor)
        anchor2 = getAnchorString(eRec.ExitAnchor)
        rule = "%spos cursive %s %s %s;%s" % (indent, tglyph, anchor1, anchor2,
                                              note)
        if context and note:
            noMatchRules.append(rule)
        else:
            rules.append(rule)

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleMarkBasePos(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    subsequentIndent = indent + INDENT

    markGlyphList = subtable.MarkCoverage.glyphs
    baseGlyphList = subtable.BaseCoverage.glyphs
    markClass = collections.defaultdict(list)
    classCount = len(set(v.Class for v in subtable.MarkArray.MarkRecord))

    for m, markGlyph in enumerate(markGlyphList):
        markRec = subtable.MarkArray.MarkRecord[m]
        markClass[markRec.Class].append((getAnchorString(markRec.MarkAnchor),
                                         markGlyph))

    # Get the mark class names from the global dict
    markClassNameList = []
    for c in markClass.keys():
        markClassName = otlConv.markClassesByLookup[otlConv.curLookupIndex,
                                                    otlConv.curSubTableIndex,
                                                    c, otlConv.markTag]
        markClassNameList.append(markClassName)

    # build a dict of [common set of anchors] -> glyph list.
    baseAnchorSets = collections.defaultdict(list)
    for b, baseGlyph in enumerate(baseGlyphList):
        baseRec = subtable.BaseArray.BaseRecord[b]
        anchorList = []
        for c in range(classCount):
            anchorList.append(getAnchorString(baseRec.BaseAnchor[c]))
        anchorKey = tuple(anchorList)
        baseAnchorSets[anchorKey].append(baseGlyph)

    for anchorKey, glyphList in baseAnchorSets.items():
        note = ""
        glyphList.sort()
        if context:
            if isinstance(inputSeqList[0], list):
                inputList = inputSeqList[0]
            else:
                inputList = [inputSeqList[0]]
            overlapList = set(glyphList).intersection(inputList)
            if not overlapList:
                note = " # Note! Not in input sequence"
                glyphTxt = " ".join(glyphList)
            else:
                glyphTxt = " ".join(sorted(overlapList))
        else:
            glyphTxt = " ".join(glyphList)

        pos_ind = '\n' + INDENT * 4
        rule = ["pos base [%s]" % pos_ind.join(gtextWrapper.wrap(glyphTxt))]

        for cl in range(classCount):
            rule.append(" %s mark %s" % (anchorKey[cl], markClassNameList[cl]))
            if (cl + 1) < classCount:  # if it is not the last one
                rule.append("\n%s" % subsequentIndent)

        rule.append(";")

        if note:
            rule.append(note)
        rule = "".join(rule)

        if context and note:
            noMatchRules.append(rule)
        else:
            rules.append(rule)

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleMarkLigPos(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    markGlyphList = subtable.MarkCoverage.glyphs
    ligGlyphList = subtable.LigatureCoverage.glyphs
    markClass = collections.defaultdict(list)
    classCount = len(set(v.Class for v in subtable.MarkArray.MarkRecord))

    for m, markGlyph in enumerate(markGlyphList):
        markRec = subtable.MarkArray.MarkRecord[m]
        markClass[markRec.Class].append((getAnchorString(markRec.MarkAnchor),
                                         markGlyph))

    # Get the mark class names from the global dict
    markClassNameList = []
    for c in markClass.keys():
        markClassName = otlConv.markClassesByLookup[otlConv.curLookupIndex,
                                                    otlConv.curSubTableIndex,
                                                    c, otlConv.markTag]
        markClassNameList.append(markClassName)

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    subsequentIndent = indent + INDENT

    for lidx, ligAttach in enumerate(subtable.LigatureArray.LigatureAttach):
        ligGlyph = ligGlyphList[lidx]
        tokenList = ["pos ligature ", ligGlyph]
        if context and (not checkGlyphInSequence(ligGlyph, inputSeqList, 0)):
            note = " # Note! Not in input sequence"
        else:
            note = ""

        for cmpIndex, componentRec in enumerate(ligAttach.ComponentRecord):
            if cmpIndex > 0:
                tokenList.append("\n%sligComponent" % (subsequentIndent * 5))
            for cl in range(classCount):
                tokenList.append(" %s mark %s" % (
                    getAnchorString(componentRec.LigatureAnchor[cl]),
                    markClassNameList[cl]))
                if (cl + 1) < classCount:
                    tokenList.append("\n%s" % subsequentIndent)

        tokenList.append(";")

        if note:
            tokenList.append(note)

        rule = "".join(tokenList)

        if context and note:
            noMatchRules.append(rule)
        else:
            rules.append(rule)

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleMarkMarkPos(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    subsequentIndent = indent + INDENT

    markGlyphList = subtable.Mark1Coverage.glyphs
    mark2GlyphList = subtable.Mark2Coverage.glyphs
    markClass = collections.defaultdict(list)
    classCount = len(set(v.Class for v in subtable.Mark1Array.MarkRecord))

    for m, markGlyph in enumerate(markGlyphList):
        markRec = subtable.Mark1Array.MarkRecord[m]
        markClass[markRec.Class].append((getAnchorString(markRec.MarkAnchor),
                                        markGlyph))

    # Get the mark class names from the global dict
    markClassNameList = []
    for c in markClass.keys():
        markClassName = otlConv.markClassesByLookup[otlConv.curLookupIndex,
                                                    otlConv.curSubTableIndex,
                                                    c, otlConv.mark1Tag]
        markClassNameList.append(markClassName)

    mark2AnchorSets = collections.defaultdict(list)
    for b, mark2Glyph in enumerate(mark2GlyphList):
        mark2Rec = subtable.Mark2Array.Mark2Record[b]
        anchorList = []
        for c in range(classCount):
            anchorList.append(getAnchorString(mark2Rec.Mark2Anchor[c]))
        anchorKey = tuple(anchorList)
        mark2AnchorSets[anchorKey].append(mark2Glyph)

    for anchorKey, glyphList in mark2AnchorSets.items():
        note = None
        # This sort will lead the rules to be sorted by base glyph names.
        glyphList.sort()
        if context:
            if isinstance(inputSeqList[0], list):
                inputList = inputSeqList[0]
            else:
                inputList = [inputSeqList[0]]
            overlapList = set(glyphList).intersection(inputList)
            if not overlapList:
                note = " # Note! Not in input sequence"
                glyphTxt = " ".join(glyphList)
            else:
                note = ""
                glyphTxt = " ".join(overlapList)
        else:
            note = ""
            glyphTxt = " ".join(glyphList)

        rule = ["%spos mark [%s]" % (indent, glyphTxt)]
        for cl in range(classCount):
            rule.append(" %s mark %s" % (anchorKey[cl], markClassNameList[cl]))
            if (cl + 1) < classCount:  # if it is not the last one
                rule.append("\n%s" % subsequentIndent)
        rule.append(";")
        if note:
            rule.append(note)
        rule = "".join(rule)
        if context and note:
            noMatchRules.append(rule)
        else:
            rules.append(rule)

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleContextPOS(subtable, otlConv, context=None):
    chainRules = []
    # pLookupList = otlConv.table.LookupList.Lookup

    # ruleContextPOS 7-3
    # XXX support is incomplete
    if subtable.Format == 3:
        log.warn('Support for GPOS LookupType 7, Format 3 is incomplete')
        inputList = []
        inputList2 = []
        for input_coverage in subtable.Coverage:
            glyphList = sorted(input_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            inputList2.append(glyphList)
            inputList.append("[" + " ".join(glyphList) + "]")
        inputTxt = "' ".join(inputList) + "'"
        rule = "pos %s;" % inputTxt

        posRules = []
        # for subRec in subtable.PosLookupRecord:
        #     lookup = pLookupList[subRec.LookupListIndex]
        #     lookupType = lookup.LookupType
        #     curLI = otlConv.curLookupIndex
        #     otlConv.curLookupIndex = subRec.LookupListIndex
        #     handler = otlConv.ruleHandlers[lookupType]
        #     contextRec = ContextRecord(inputList2, subRec.SequenceIndex)

        #     for si, sub_table in enumerate(lookup.SubTable):
        #         curSI = otlConv.curSubTableIndex
        #         otlConv.curSubTableIndex = si
        #         subtablerules = handler(sub_table, otlConv, contextRec)
        #         otlConv.curSubTableIndex = curSI
        #         posRules.extend(subtablerules)

        #     otlConv.curLookupIndex = curLI

        chainRules.append([rule, posRules])

    # ruleContextPOS 7-2
    # XXX support is incomplete
    elif subtable.Format == 2:
        log.warn('Support for GPOS LookupType 7, Format 2 is incomplete')
        for i, ctxClassSet in enumerate(subtable.PosClassSet):
            if not ctxClassSet:
                continue

            for ctxClassRule in ctxClassSet.PosClassRule:
                className = otlConv.classesByLookup[
                    otlConv.curLookupIndex, otlConv.curSubTableIndex, i, None]
                inputList = [className]
                inputList2 = [otlConv.classesByClassName[className]]

                for classIndex in ctxClassRule.Class:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, None]
                    inputList.append(className)
                    inputList2.append(otlConv.classesByClassName[className])
                inputTxt = "' ".join(inputList) + "'"
                rule = "pos %s;" % inputTxt

                posRules = []
                # for subsRec in ctxClassRule.PosLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(
                #         inputList, subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         posRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, posRules])

    # ruleContextPOS 7-1
    # XXX support is incomplete
    elif subtable.Format == 1:
        log.warn('Support for GPOS LookupType 7, Format 1 is incomplete')
        firstGlyphList = subtable.Coverage.glyphs
        # for each glyph in the coverage table
        for ri, subRuleSet in enumerate(subtable.PosRuleSet):
            firstGlyph = firstGlyphList[ri]

            for ctxRuleRec in subRuleSet.PosRule:
                inputList = [firstGlyph] + ctxRuleRec.Input
                inputTxt = "' ".join(inputList) + "'"
                rule = "pos %s" % inputTxt

                posRules = []
                # for subsRec in ctxRuleRec.PosLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(inputList,
                #                                subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         posRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, posRules])
    else:
        raise AttributeError("Unknown Context Pos format %s" %
                             subtable.Format)

    rules = []
    for chainRule, posRules in sorted(chainRules):
        rules.append(chainRule)
        rules.extend(posRules)
    return rules


def ruleChainContextPOS(subtable, otlConv, context=None):
    chainRules = []
    # pLookupList = otlConv.table.LookupList.Lookup

    # ruleChainContextPOS 8-3
    # XXX support is incomplete
    if subtable.Format == 3:
        log.warn('Support for GPOS LookupType 8, Format 3 is incomplete')
        backtrackList = []
        for backtrack_coverage in subtable.BacktrackCoverage:
            glyphList = sorted(backtrack_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            backtrackList.append("[" + " ".join(glyphList) + "]")
        backtrackList.reverse()
        backTxt = " ".join(backtrackList)

        inputList = []
        inputList2 = []
        for input_coverage in subtable.InputCoverage:
            glyphList = sorted(input_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            inputList2.append(glyphList)
            inputList.append("[" + " ".join(glyphList) + "]")
        inputTxt = "' ".join(inputList) + "'"

        lookAheadList = []
        for lookahead_coverage in subtable.LookAheadCoverage:
            glyphList = sorted(lookahead_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            lookAheadList.append("[" + " ".join(glyphList) + "]")
        lookAheadTxt = " ".join(lookAheadList)
        rule = "pos %s %s %s;" % (backTxt, inputTxt, lookAheadTxt)

        posRules = []
        # for subRec in subtable.PosLookupRecord:
        #     lookup = pLookupList[subRec.LookupListIndex]
        #     lookupType = lookup.LookupType
        #     curLI = otlConv.curLookupIndex
        #     otlConv.curLookupIndex = subRec.LookupListIndex
        #     handler = otlConv.ruleHandlers[lookupType]
        #     contextRec = ContextRecord(inputList2, subRec.SequenceIndex)

        #     for si, sub_table in enumerate(lookup.SubTable):
        #         curSI = otlConv.curSubTableIndex
        #         otlConv.curSubTableIndex = si
        #         subtablerules = handler(sub_table, otlConv, contextRec)
        #         otlConv.curSubTableIndex = curSI
        #         posRules.extend(subtablerules)

        #     otlConv.curLookupIndex = curLI

        chainRules.append([rule, posRules])

    # ruleChainContextPOS 8-2
    # XXX support is incomplete
    elif subtable.Format == 2:
        log.warn('Support for GPOS LookupType 8, Format 2 is incomplete')
        for i, ctxClassSet in enumerate(subtable.ChainSubClassSet):
            if not ctxClassSet:
                continue

            for ctxClassRule in ctxClassSet.ChainPosRuleSet:
                backTrackList = []
                for classIndex in ctxClassRule.Backtrack:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, otlConv.backtrackTag]
                    backTrackList.append(className)
                backTxt = " ".join(backTrackList)

                className = otlConv.classesByLookup[
                    otlConv.curLookupIndex, otlConv.curSubTableIndex, i,
                    otlConv.InputTag]
                inputList = [className]
                inputList2 = [otlConv.classesByClassName[className]]

                for classIndex in ctxClassRule.Input:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, otlConv.InputTag]
                    inputList.append(className)
                    inputList2.append(otlConv.classesByClassName[className])
                inputTxt = "' ".join(inputList) + "'"

                lookAheadList = []
                for classIndex in ctxClassRule.LookAhead:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, otlConv.lookAheadTag]
                    lookAheadList.append(className)
                lookTxt = " ".join(lookAheadList)
                rule = "sub %s %s %s;" % (backTxt, inputTxt, lookTxt)

                posRules = []
                # for subsRec in ctxClassRule.SubstLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(inputList,
                #                                subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         posRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, posRules])

    # ruleChainContextPOS 8-1
    # XXX support is incomplete
    elif subtable.Format == 1:
        log.warn('Support for GPOS LookupType 8, Format 1 is incomplete')
        firstGlyphList = subtable.Coverage.glyphs
        # for each glyph in the coverage table
        for ri, subRuleSet in enumerate(subtable.ChainPosRuleSet):
            firstGlyph = firstGlyphList[ri]

            for ctxRuleRec in subRuleSet.ChainPosRule:
                backList = ctxRuleRec.Backtrack
                backList.reverse()
                backTxt = " ".join(backList)
                inputList = [firstGlyph] + ctxRuleRec.Input
                inputTxt = "' ".join(inputList) + "'"
                lookAheadTxt = " ".join(ctxRuleRec.LookAhead)
                rule = "sub %s %s %s" % (backTxt, inputTxt, lookAheadTxt)

                posRules = []
                # for subsRec in ctxRuleRec.SubstLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(inputList,
                #                                subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         posRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, posRules])
    else:
        raise AttributeError("Unknown Chain Context Pos format %s" % (
            subtable.Format))

    rules = []
    for chainRule, posRules in sorted(chainRules):
        rules.append(chainRule)
        rules.extend(posRules)
    return rules


def ruleExt(subtable, otlConv, context=None):
    handler = otlConv.ruleHandlers[subtable.ExtensionLookupType]
    rules = handler(subtable.ExtSubTable, otlConv, context)
    return rules


def ruleSingleSub(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    for g1, g2 in subtable.mapping.items():
        rule = note = None
        if context and (not checkGlyphInSequence(g1, inputSeqList, 0)):
            note = " # Note! Not in input sequence"
        else:
            note = ""
        rule = "%ssub %s by %s;%s" % (indent, g1, g2, note)
        if context and note:
            noMatchRules.append(rule)
        else:
            rules.append(rule)

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleMultipleSub(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    for g1, substGlyphList in subtable.mapping.items():
        rule = note = None
        if context and (not checkGlyphInSequence(g1, inputSeqList, 0)):
            note = " # Note! Not in input sequence"
        else:
            note = ""
        subTxt = " ".join(substGlyphList)
        rule = "%ssub %s by [%s];%s" % (indent, g1, subTxt, note)
        if context and note:
            noMatchRules.append(rule)
        else:
            rules.append(rule)

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleAltSub(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context
    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    for g1, alts in subtable.alternates.items():
        rule = note = None
        if context and (not checkGlyphInSequence(g1, inputSeqList, 0)):
            note = " # Note! Not in input sequence"
        else:
            note = ""
        altText = " ".join(sorted(alts))
        rule = "%ssub %s from [%s];%s" % (indent, g1, altText, note)
        if context and note:
            noMatchRules.append(rule)
        else:
            rules.append(rule)

    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleLigatureSub(subtable, otlConv, context=None):
    rules = []
    noMatchRules = []  # contains the rules that don't match the context

    if context:
        indent = ChainINDENT
        inputSeqList = context.inputList[context.sequenceIndex:]
    else:
        indent = ""
        inputSeqList = []

    for item in subtable.ligatures.items():
        ligs = item[1]
        firstGlyph = item[0]
        for lig in ligs:
            rule = note = None
            tokenList = [firstGlyph]
            missing = []
            foundAll = 1
            if context:
                try:
                    if not checkGlyphInSequence(firstGlyph, inputSeqList, 0):
                        raise ValueError

                    # make sure everything else is also present with
                    # the same index.
                    for i, gname in enumerate(lig.Component):
                        if not checkGlyphInSequence(gname, inputSeqList,
                                                    i + 1):
                            foundAll = 0
                            missing.append(gname)

                    if not foundAll:
                        note = (" # Note! lig components %s are not in input "
                                "sequence with same index."
                                % " ".join(missing))
                    else:
                        note = ""
                except ValueError:
                    note = (" # Note! first glyph %s is not in input "
                            "sequence." % firstGlyph)
            else:
                note = ""

            tokenList.extend(lig.Component)
            glyphTxt = " ".join(tokenList)
            rule = "%ssub %s by %s;%s" % (indent, glyphTxt, lig.LigGlyph, note)
            if context and note:
                noMatchRules.append(rule)
            else:
                rules.append(rule)
    if rules:
        return sorted(rules)
    else:
        return sorted(noMatchRules)


def ruleContextSUB(subtable, otlConv, context=None):
    chainRules = []
    # pLookupList = otlConv.table.LookupList.Lookup

    # ruleContextSUB 5-3
    # XXX support is incomplete
    if subtable.Format == 3:
        log.warn('Support for GSUB LookupType 5, Format 3 is incomplete')
        inputList = []
        inputList2 = []

        for input_coverage in subtable.Coverage:
            glyphList = sorted(input_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            inputList2.append(glyphList)
            inputList.append("[" + " ".join(glyphList) + "]'")
        inputTxt = "' ".join(inputList) + "'"
        rule = "sub %s;" % inputTxt

        subRules = []
        # for subRec in subtable.SubstLookupRecord:
        #     lookup = pLookupList[subRec.LookupListIndex]
        #     lookupType = lookup.LookupType
        #     curLI = otlConv.curLookupIndex
        #     otlConv.curLookupIndex = subRec.LookupListIndex
        #     handler = otlConv.ruleHandlers[lookupType]
        #     contextRec = ContextRecord(inputList2, subRec.SequenceIndex)

        #     for si, sub_table in enumerate(lookup.SubTable):
        #         curSI = otlConv.curSubTableIndex
        #         otlConv.curSubTableIndex = si
        #         subtablerules = handler(sub_table, otlConv, contextRec)
        #         otlConv.curSubTableIndex = curSI
        #         subRules.extend(subtablerules)

        #     otlConv.curLookupIndex = curLI

        chainRules.append([rule, subRules])

    # ruleContextSUB 5-2
    # XXX support is incomplete
    elif subtable.Format == 2:
        log.warn('Support for GSUB LookupType 5, Format 2 is incomplete')
        for i, ctxClassSet in enumerate(subtable.SubClassSet):
            if not ctxClassSet:
                continue

            for ctxClassRule in ctxClassSet.SubClassRule:
                className = otlConv.classesByLookup[
                    otlConv.curLookupIndex, otlConv.curSubTableIndex, i, None]
                inputList = [className]
                inputList2 = [otlConv.classesByClassName[className]]

                for classIndex in ctxClassRule.Class:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, None]
                    inputList.append(className)
                    inputList2.append(otlConv.classesByClassName[className])
                inputTxt = "' ".join(inputList) + "'"
                rule = "sub %s;" % inputTxt

                subRules = []
                # for subsRec in ctxClassRule.SubstLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(
                #         inputList, subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         subRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, subRules])

    # ruleContextSUB 5-1
    # XXX support is incomplete
    elif subtable.Format == 1:
        log.warn('Support for GSUB LookupType 5, Format 1 is incomplete')
        firstGlyphList = subtable.Coverage.glyphs
        # for each glyph in the coverage table
        for ri, subRuleSet in enumerate(subtable.SubRuleSet):
            firstGlyph = firstGlyphList[ri]

            for ctxRuleRec in subRuleSet.SubRule:
                inputList = [firstGlyph] + ctxRuleRec.Input
                inputTxt = "' ".join(inputList) + "'"
                rule = "sub %s" % inputTxt

                subRules = []
                # for subsRec in ctxRuleRec.SubstLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(inputList,
                #                                subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         subRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, subRules])
    else:
        raise AttributeError("Unknown Context Sub format %s" % (
            subtable.Format))

    rules = []
    for chainRule, subRules in sorted(chainRules):
        rules.append(chainRule)
        rules.extend(subRules)
    return rules


def wrap_statement(token_str):
    """
    Wraps a long string of space-separated tokens
    or a list of tokens.
    """
    if isinstance(token_str, list):
        token_str = ' '.join(token_str)
    wrap_ind = '\n' + INDENT * 4
    return wrap_ind.join(gtextWrapper.wrap(token_str))


def ruleChainContextSUB(subtable, otlConv, context=None):
    chainRules = []
    # pLookupList = otlConv.table.LookupList.Lookup

    # ruleChainContextSUB 6-3
    # XXX support is incomplete
    if subtable.Format == 3:
        log.warn('Support for GSUB LookupType 6, Format 3 is incomplete')
        backtrackList = []
        for backtrack_coverage in subtable.BacktrackCoverage:
            glyphList = sorted(backtrack_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]

            backtrackList.append("[" + " ".join(glyphList) + "]")
        backtrackList.reverse()
        backTxt = " ".join(backtrackList)

        inputList = []
        inputList2 = []
        for input_coverage in subtable.InputCoverage:
            glyphList = sorted(input_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            inputList2.append(glyphList)
            inputList.append("[" + " ".join(glyphList) + "]'")
        inputTxt = " ".join(inputList)

        lookAheadList = []
        for lookahead_coverage in subtable.LookAheadCoverage:
            glyphList = sorted(lookahead_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            lookAheadList.append("[" + " ".join(glyphList) + "]")
        lookAheadTxt = " ".join(lookAheadList)
        rule = wrap_statement(
            "sub %s %s %s;" % (backTxt, inputTxt, lookAheadTxt))

        subRules = []
        # for subsRec in subtable.SubstLookupRecord:
        #     lookup = pLookupList[subsRec.LookupListIndex]
        #     lookupType = lookup.LookupType
        #     curLI = otlConv.curLookupIndex
        #     otlConv.curLookupIndex = subsRec.LookupListIndex
        #     handler = otlConv.ruleHandlers[lookupType]
        #     contextRec = ContextRecord(inputList2, subsRec.SequenceIndex)

        #     for si, sub_table in enumerate(lookup.SubTable):
        #         curSI = otlConv.curSubTableIndex
        #         otlConv.curSubTableIndex = si
        #         subtablerules = handler(sub_table, otlConv, contextRec)
        #         otlConv.curSubTableIndex = curSI
        #         subRules.extend(subtablerules)

        #     otlConv.curLookupIndex = curLI

        chainRules.append([rule, subRules])

    # ruleChainContextSUB 6-2
    # XXX support is incomplete
    elif subtable.Format == 2:
        log.warn('Support for GSUB LookupType 6, Format 2 is incomplete')
        for i, ctxClassSet in enumerate(subtable.ChainSubClassSet):
            if not ctxClassSet:
                continue

            for ctxClassRule in ctxClassSet.ChainSubClassRule:
                backTrackList = []

                for classIndex in ctxClassRule.Backtrack:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, otlConv.backtrackTag]
                    backTrackList.append(className)
                backTxt = " ".join(backTrackList)

                className = otlConv.classesByLookup[
                    otlConv.curLookupIndex, otlConv.curSubTableIndex, i,
                    otlConv.InputTag]
                inputList = [className]
                inputList2 = [otlConv.classesByClassName[className]]

                for classIndex in ctxClassRule.Input:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, otlConv.InputTag]
                    inputList.append(className)
                    inputList2.append(otlConv.classesByClassName[className])
                inputTxt = " ".join(inputList)

                lookAheadList = []
                for classIndex in ctxClassRule.LookAhead:
                    className = otlConv.classesByLookup[
                        otlConv.curLookupIndex, otlConv.curSubTableIndex,
                        classIndex, otlConv.lookAheadTag]
                    lookAheadList.append(className)
                lookTxt = " ".join(lookAheadList)
                rule = "sub %s %s' %s;" % (backTxt, inputTxt, lookTxt)

                subRules = []
                # for subsRec in ctxClassRule.SubstLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(
                #         inputList, subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         subRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, subRules])

    # ruleChainContextSUB 6-1
    # XXX support is incomplete
    elif subtable.Format == 1:
        log.warn('Support for GSUB LookupType 6, Format 1 is incomplete')
        firstGlyphList = subtable.Coverage.glyphs
        # for each glyph in the coverage table
        for ri, subRuleSet in enumerate(subtable.ChainSubRuleSet):
            firstGlyph = firstGlyphList[ri]

            for ctxRuleRec in subRuleSet.ChainSubRule:
                backList = ctxRuleRec.Backtrack
                backList.reverse()
                backTxt = " ".join(backList)
                inputList = [firstGlyph] + ctxRuleRec.Input
                inputTxt = " ".join(inputList)
                lookAheadTxt = " ".join(ctxRuleRec.LookAhead)
                rule = "sub %s %s' %s by " % (backTxt, inputTxt, lookAheadTxt)

                subRules = []
                # for subsRec in ctxRuleRec.SubstLookupRecord:
                #     lookup = pLookupList[subsRec.LookupListIndex]
                #     lookupType = lookup.LookupType
                #     curLI = otlConv.curLookupIndex
                #     otlConv.curLookupIndex = subsRec.LookupListIndex
                #     handler = otlConv.ruleHandlers[lookupType]
                #     contextRec = ContextRecord(inputList,
                #                                subsRec.SequenceIndex)

                #     for si, sub_table in enumerate(lookup.SubTable):
                #         curSI = otlConv.curSubTableIndex
                #         otlConv.curSubTableIndex = si
                #         subtablerules = handler(
                #             sub_table, otlConv, contextRec)
                #         otlConv.curSubTableIndex = curSI
                #         subRules.extend(subtablerules)

                #     otlConv.curLookupIndex = curLI

                chainRules.append([rule, subRules])
    else:
        raise AttributeError("Unknown Chain Context Sub format %s" % (
            subtable.Format))

    rules = []
    for chainRule, subRules in sorted(chainRules):
        rules.append(chainRule)
        rules.extend(subRules)
    return rules


def ruleReverseChainSub(subtable, otlConv, context=None):
    rules = []
    if subtable.Format == 1:
        backtrackList = []
        for backtrack_coverage in subtable.BacktrackCoverage:
            glyphList = sorted(backtrack_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            backtrackTxt = "[ %s ] " % " ".join(glyphList)
            backtrackList.append(backtrackTxt)
        backtrackList.reverse()
        backTxt = " ".join(backtrackList)

        glyphList = subtable.Coverage.glyphs
        # Since the substitute list is in the same order as the
        # coverage table, I need to sort the substitution array
        # in the same order as the input list.

        glyphList = sorted([[glyph, i] for i, glyph in enumerate(glyphList)])
        substituteList = []
        if subtable.Substitute:
            for entry in glyphList:
                substituteList.append(subtable.Substitute[entry[1]])

        for i in reversed(range(1, len(glyphList))):
            if glyphList[i - 1] == glyphList[i]:
                del glyphList[i]
                del substituteList[i]

        glyphList = [entry[0] for entry in glyphList]

        inputTxt = "[ %s ]" % " ".join(glyphList)

        lookAheadList = []
        for lookahead_coverage in subtable.LookAheadCoverage:
            glyphList = sorted(lookahead_coverage.glyphs)
            for i in reversed(range(1, len(glyphList))):
                if glyphList[i - 1] == glyphList[i]:
                    del glyphList[i]
            lookAheadTxt = " [ %s ]" % " ".join(glyphList)
            lookAheadList.append(lookAheadTxt)
        lookAheadTxt = " ".join(lookAheadList)

        if subtable.Substitute:
            replText = " by [ %s ]" % (" ".join(substituteList))
        else:
            replText = ""
        rule = "sub %s%s'%s%s;" % (backTxt, inputTxt, lookAheadTxt, replText)
        rules.append(rule)
    return rules


def getValRec(valueFormat, valueRec):
    if valueFormat == 4:
        return "%d" % valueRec.XAdvance
    else:
        xPos = yPos = xAdv = yAdv = 0
        if valueFormat & 1:
            xPos = valueRec.XPlacement
        if valueFormat & 2:
            yPos = valueRec.YPlacement
        if valueFormat & 4:
            xAdv = valueRec.XAdvance
        if valueFormat & 8:
            yAdv = valueRec.YAdvance

        if xPos == yPos == xAdv == yAdv == 0:
            return "0"

        return getValueRecord(xPos, yPos, xAdv, yAdv)


gLookupFlagMap = {
    1: "RightToLeft",
    2: "IgnoreBaseGlyphs",
    4: "IgnoreLigatures",
    8: "IgnoreMarks",
    0xFF00: "MarkAttachmentTypedef",
}

gLookupFlagKeys = sorted(gLookupFlagMap.keys())


def getLookupFlagTag(lookupFlag):
    retList = []
    for lookupVal in gLookupFlagKeys:
        if lookupVal & lookupFlag:
            retList.append(gLookupFlagMap[lookupVal])
            if lookupVal == 0xFF00:
                classIndex = (lookupVal & lookupFlag) >> 8
                retList.append("%d" % classIndex)
    return " ".join(retList)


DFLT_LANGSYS_KEYS = (("DFLT", "dflt"), ("dflt", "dflt"))


class OTLConverter(object):
    leftSideTag = "Left"
    rightSideTag = "Right"
    backtrackTag = "BacktrackClassDef"
    InputTag = "InputClassDef"
    lookAheadTag = "LookAheadClassDef"
    markTag = "MarkClass"
    mark1Tag = "MarkMarkClass1"
    mark2Tag = "MarkMarkClass2"

    def __init__(self, writer, ttFont, table_tag):
        self.writer = writer
        self.table = ttFont[table_tag].table
        self.classesByNameList = collections.defaultdict(list)
        self.classesByLookup = {}
        self.classesByClassName = {}
        self.lookupIndex = -1
        self.seenLookup = {}
        self.glyphDict = {}
        self.markClassesByDefList = collections.defaultdict(list)
        self.markClassesByLookup = collections.defaultdict(list)
        self.markClassesByClassName = {}
        self.curLookupIndex = -1
        self.curSubTableIndex = -1
        self.curLookupName = ""
        self.showExtension = ttFont.showExtensionFlag

        for name in ttFont.getGlyphOrder():
            self.glyphDict[name] = 0

        if table_tag == "GPOS":
            self.ExtensionIndex = 9
            self.classHandlers = {
                (2, 2): classPairGPOS,
                (4, 1): markClassPOS,
                (5, 1): markLigClassPOS,
                (6, 1): markMarkClassPOS,
                (7, 2): classContext,
                (8, 2): classChainContext,
                (9, 1): classExt,
            }
            self.ruleHandlers = {
                1: ruleSinglePos,
                2: rulePairPos,
                3: ruleCursivePos,
                4: ruleMarkBasePos,
                5: ruleMarkLigPos,
                6: ruleMarkMarkPos,
                7: ruleContextPOS,
                8: ruleChainContextPOS,
                9: ruleExt,
            }

        elif table_tag == "GSUB":
            self.ExtensionIndex = 7
            self.classHandlers = {
                (5, 2): classContext,
                (6, 2): classChainContext,
                (7, 1): classExt,
            }
            self.ruleHandlers = {
                1: ruleSingleSub,
                2: ruleMultipleSub,
                3: ruleAltSub,
                4: ruleLigatureSub,
                5: ruleContextSUB,
                6: ruleChainContextSUB,
                7: ruleExt,
                8: ruleReverseChainSub,
            }
        else:
            raise KeyError("OTLConverter can only be called for GPOS and "
                           "GSUB tables")

    def otlFeatureFormat(self):
        # get and write language systems
        lsList, featDictByLangSys, featDictByIndex = self.buildLangSys()
        self.writeLangSys(lsList)

        self.writer.newline()

        # get and write class defs.
        self.buildClasses()
        self.writeClasses()
        self.writeMarkClasses()

        # do feature defs
        self.doFeatures(featDictByLangSys)

    @staticmethod
    def addLangSysEntry(pfeatList, scriptTag, langTag, langSys,
                        featDictByLangSys, featDictByIndex):
        for featIndex in langSys.FeatureIndex:
            try:
                featRecord = pfeatList[featIndex]
            except IndexError:
                log.error("FeatList does not contain index %s from current "
                          "langsys (%s, %s).", featIndex, scriptTag, langTag)
                log.error("FeaturListLen %s, langsys FeatIndex %s.",
                          len(pfeatList), langSys.FeatureIndex)
                continue
            featTag = featRecord.FeatureTag
            lang_sys = (scriptTag, langTag)
            try:
                featDictByIndex[featIndex].append(lang_sys)
            except KeyError:
                featDictByIndex[featIndex] = [lang_sys]

            langsysDict = featDictByLangSys.get(featTag, {})
            try:
                langsysDict[lang_sys].append(featIndex)
            except KeyError:
                langsysDict[lang_sys] = [featIndex]
            featDictByLangSys[featTag] = langsysDict

    def buildLangSys(self):
        pfeatList = self.table.FeatureList.FeatureRecord
        featDictByLangSys = {}
        featDictByIndex = {}
        haveDFLT = False
        lsList = []
        for scriptRecord in self.table.ScriptList.ScriptRecord:
            scriptTag = scriptRecord.ScriptTag
            if scriptTag in ("DFLT", "dflt"):
                haveDFLT = True
            if scriptRecord.Script.DefaultLangSys:
                langTag = "dflt"
                lsList.insert(0, (
                    scriptTag, langTag,
                    scriptRecord.Script.DefaultLangSys.ReqFeatureIndex))
                self.addLangSysEntry(pfeatList, scriptTag, langTag,
                                     scriptRecord.Script.DefaultLangSys,
                                     featDictByLangSys, featDictByIndex)

            for langSysRecord in scriptRecord.Script.LangSysRecord:
                langTag = langSysRecord.LangSysTag
                lsList.append((scriptTag, langTag,
                              langSysRecord.LangSys.ReqFeatureIndex))
                self.addLangSysEntry(pfeatList, scriptTag, langTag,
                                     langSysRecord.LangSys, featDictByLangSys,
                                     featDictByIndex)

        if not haveDFLT:
            # Synthesize a DFLT dflt entry.
            addedDFLT = False
            scriptCount = len(self.table.ScriptList.ScriptRecord)
            for langsysList in featDictByIndex.values():
                if len(langsysList) == scriptCount:
                    langsysList.insert(0, ("DFLT", "dflt", 0xFFFF))
                    addedDFLT = True

            if addedDFLT:
                lsList.insert(0, ("DFLT", "dflt", 0xFFFF))

        return lsList, featDictByLangSys, featDictByIndex

    def writeLangSys(self, lsList):
        # lsList is alphabetic, because of font order, except that
        # I moved (DFLT, dflt) script to the top.
        for lang_sys in lsList:
            if lang_sys[2] != 0xFFFF:
                rfTxt = "  Required Feature Index %s" % (lang_sys[2])
            else:
                rfTxt = ""
            self.writer.write("languagesystem %s %s%s;" % (
                lang_sys[0], lang_sys[1], rfTxt))
            self.writer.newline()

    def buildClasses(self):
        for lookupIndex, lookup in enumerate(self.table.LookupList.Lookup):
            self.curLookupIndex = lookupIndex
            for subtableIndex, subtable in enumerate(lookup.SubTable):
                self.curSubTableIndex = subtableIndex
                handler = self.classHandlers.get(
                    (lookup.LookupType, subtable.Format))
                if handler:
                    handler(subtable, self)

        for nameList in sorted(self.classesByNameList.keys()):
            classRecList = self.classesByNameList[nameList]
            lenList = len(nameList)
            if lenList == 0:
                className = "@empty"
            elif lenList == 1:
                className = "[%s]" % (nameList[0])
            elif lenList == 2:
                className = "@%s_%s" % (nameList[0], nameList[1])
            else:
                className = "@%s_%d_%s" % (nameList[0], len(nameList),
                                           nameList[-1])
                i = 1
                while (i < lenList) and className in self.classesByClassName:
                    className = "@%s_%s_%d_%s" % (nameList[0], nameList[i],
                                                  len(nameList), nameList[-1])
            self.classesByClassName[className] = nameList
            for classRec in classRecList:
                key = (classRec.lookupIndex, classRec.subtableIndex,
                       classRec.classIndex, classRec.side)
                self.classesByLookup[key] = className

        for defList in sorted(self.markClassesByDefList.keys()):
            classRecList = self.markClassesByDefList[defList]
            defNameList = []
            for nameList, _anchor in defList:
                defNameList.extend(list(nameList))
            lenList = len(defNameList)
            if lenList == 0:
                className = "@empty_mark"
            elif lenList == 1:
                className = "%s_mark" % (defNameList[0])
            elif lenList == 2:
                className = "@%s_%s_mark" % (defNameList[0], defNameList[1])
            else:
                className = "@%s_%d_%s_mark" % (
                    defNameList[0], len(defNameList), defNameList[-1])
                i = 1
                while (i < lenList) and className in self.classesByClassName:
                    className = "@%s_%s_%d_%s_mark" % (
                        defNameList[0], defNameList[i], len(defNameList),
                        defNameList[-1])

            self.markClassesByClassName[className] = defList

            for classRec in classRecList:
                key = (classRec.lookupIndex, classRec.subtableIndex,
                       classRec.classIndex, classRec.side)
                self.markClassesByLookup[key] = className

    def writeClasses(self):
        classNames = list(self.classesByClassName.keys())
        if classNames:
            self.writer.newline()
            self.writer.write(
                "# Class definitions *********************************")
            self.writer.newline()
        for className in sorted(classNames):
            if className[0] == "[":
                # we don't write out the single glyph class names,
                # as they are ued in -line as class lists.
                continue
            class_glyphs = " ".join(self.classesByClassName[className])
            class_statement = "%s = [ %s ];" % (className, class_glyphs)
            class_lines = gtextWrapper.wrap(class_statement)
            for wrapped_line in class_lines:
                self.writer._writeraw(wrapped_line)
                self.writer.newline()
        self.writer.newline()

    def writeMarkClasses(self):
        classNames = list(self.markClassesByClassName.keys())
        if classNames:
            self.writer.newline()
            self.writer.write(
                "# Mark Class definitions *********************************")
            self.writer.newline()
        for className in sorted(classNames):
            classDef = self.markClassesByClassName[className]
            for gnameList, anchor in classDef:
                self.writer._writeraw("mark [")
                class_lines = gtextWrapper.wrap(" ".join(gnameList))
                num_lines = len(class_lines)
                for ln, wrapped_line in enumerate(class_lines):
                    self.writer._writeraw(wrapped_line)
                    if ln < num_lines - 1:
                        self.writer.newline()
                self.writer._writeraw("] %s %s;" % (anchor, className))
                self.writer.newline()
        self.writer.newline()

    def doFeatures(self, featDictByLangSys):
        for featTag in sorted(featDictByLangSys.keys()):
            self.writer.write("feature %s {" % featTag)
            self.writer.newline()
            self.writer.indent()
            langSysDict = featDictByLangSys[featTag]
            langSysTagList = sorted(langSysDict.keys())

            dflt_langsys_key = None
            dfltFeatIndexList = None
            for key in DFLT_LANGSYS_KEYS:
                if key in langSysDict:
                    dfltFeatIndexList = langSysDict[key]
                    langSysTagList.remove(key)
                    dflt_langsys_key = key
                    break

            dfltLookupIndexDict = None
            if dfltFeatIndexList:
                dfltLookupIndexDict = self.writeDfltLangSysFeat(
                    dflt_langsys_key, dfltFeatIndexList)

            if dflt_langsys_key:
                prevLangSysKey = dflt_langsys_key
            else:
                prevLangSysKey = DFLT_LANGSYS_KEYS[0]

            for langSysKey in langSysTagList:
                self.writeLangSysFeat(langSysKey, prevLangSysKey,
                                      langSysDict[langSysKey],
                                      dfltLookupIndexDict)
                prevLangSysKey = langSysKey

            self.writer.dedent()
            self.writer.write("} %s;" % featTag)
            self.writer.newline()
            self.writer.newline()

    def writeDfltLangSysFeat(self, langSysKey, featIndexList):
        # XXX similar to writeLangSysFeat()
        pfeatList = self.table.FeatureList.FeatureRecord
        pLookupList = self.table.LookupList.Lookup
        featRecord = None

        lookupIndexDict = {}
        # key: lookup index (int)
        # val: fontTools.ttLib.tables.otTables.Lookup (object)
        for featIndex in featIndexList:
            featRecord = pfeatList[featIndex]
            for lookupIndex in featRecord.Feature.LookupListIndex:
                lookupIndexDict[lookupIndex] = pLookupList[lookupIndex]

        excludeDFLTtxt = ""
        nameIndex = 0
        for li in sorted(lookupIndexDict.keys()):
            self.curLookupIndex = li
            lookup = pLookupList[li]
            lookupFlagTxt = getLookupFlagTag(lookup.LookupFlag)

            useExtension = ""
            if self.showExtension and lookup.LookupType == self.ExtensionIndex:
                useExtension = " useExtension"

            if li in self.seenLookup:
                lookupName = self.seenLookup[li]
                self.writer.write("lookup %s%s;" % (
                    lookupName, excludeDFLTtxt))
                self.writer.newline()
            else:
                lookupName = "%s_%s_%s_%s" % (
                    featRecord.FeatureTag, langSysKey[0], langSysKey[1],
                    nameIndex)
                self.seenLookup[li] = lookupName
                self.writer.write("lookup %s%s {" % (lookupName,
                                                     useExtension))
                excludeDFLTtxt = ""  # Only need to write it once.

                self.writer.newline()
                self.writer.indent()

                if lookupFlagTxt:
                    self.writer.write("lookupflag %s;" % lookupFlagTxt)
                    self.writer.newline()

                self.curLookupIndex = li
                self.writeLookup(lookup)
                self.writer.dedent()
                self.writer.write("} %s;" % lookupName)
                self.writer.newline()
                self.writer.newline()
                nameIndex += 1

        return lookupIndexDict

    def writeLangSysFeat(self, langSysKey, prevLangSysKey,
                         featIndexList, dfltLookupIndexDict):
        pfeatList = self.table.FeatureList.FeatureRecord
        pLookupList = self.table.LookupList.Lookup
        featRecord = None

        lookupIndexDict = {}
        # key: lookup index (int)
        # val: fontTools.ttLib.tables.otTables.Lookup (object)
        for featIndex in featIndexList:
            featRecord = pfeatList[featIndex]
            for lookupIndex in featRecord.Feature.LookupListIndex:
                lookupIndexDict[lookupIndex] = pLookupList[lookupIndex]

        # Remove all lookups shared with the DFLT/dflt script.
        # Note if there were any; if not, then we need to use the
        # exclude keyword with the lookup.
        haveAnyDflt = False
        excludeDFLT = False
        if dfltLookupIndexDict:
            for lookupIndex in sorted(lookupIndexDict.keys()):
                if lookupIndex in dfltLookupIndexDict:
                    del lookupIndexDict[lookupIndex]
                    haveAnyDflt = True
            if not haveAnyDflt:
                excludeDFLT = True
        liList = sorted(lookupIndexDict.keys())

        excludeDFLTtxt = ""
        if excludeDFLT:
            excludeDFLTtxt = " excludeDFLT"

        # If all the lookups were shared with DFLt dflt,
        # no need to write anything.
        if liList:
            nameIndex = 0
            if prevLangSysKey[0] != langSysKey[0]:
                self.writer.write("script %s;" % (langSysKey[0]))
                self.writer.newline()
                self.writer.write("language %s;" % (langSysKey[1]))
                self.writer.newline()
            elif prevLangSysKey[1] != langSysKey[1]:
                self.writer.write("language %s;" % (langSysKey[1]))
                self.writer.newline()

            for li in liList:
                self.curLookupIndex = li
                lookup = pLookupList[li]
                lookupFlagTxt = getLookupFlagTag(lookup.LookupFlag)

                useExtension = ""
                if self.showExtension and (
                        lookup.LookupType == self.ExtensionIndex):
                    useExtension = " useExtension"

                if li in self.seenLookup:
                    lookupName = self.seenLookup[li]
                    self.writer.write("lookup %s%s;" % (
                        lookupName, excludeDFLTtxt))
                    excludeDFLTtxt = ""  # Only need to write it once.
                    self.writer.newline()
                else:
                    lookupName = "%s_%s_%s_%s" % (
                        featRecord.FeatureTag, langSysKey[0].strip(),
                        langSysKey[1].strip(), nameIndex)
                    self.seenLookup[li] = lookupName
                    self.writer.write("lookup %s%s {" % (lookupName,
                                                         useExtension))
                    excludeDFLTtxt = ""  # Only need to write it once.

                    self.writer.newline()
                    self.writer.indent()

                    if lookupFlagTxt:
                        self.writer.write("lookupflag %s;" % lookupFlagTxt)
                        self.writer.newline()

                    self.curLookupIndex = li
                    self.writeLookup(lookup)
                    self.writer.dedent()
                    self.writer.write("} %s;" % lookupName)
                    self.writer.newline()
                    self.writer.newline()
                    nameIndex += 1

    def writeLookup(self, lookup):
        lookupType = lookup.LookupType
        handler = self.ruleHandlers[lookupType]

        for si, sub_table in enumerate(lookup.SubTable):
            self.curSubTableIndex = si
            rules = handler(sub_table, self)
            for rule in rules:
                self.writer._writeraw(rule)
                self.writer.newline()


def dumpOTLAsFeatureFile(writer, ttFont, table_tag):
    otlConverter = OTLConverter(writer, ttFont, table_tag)
    otlConverter.otlFeatureFormat()


class TTXNTTFont(TTFont):
    def __init__(self, file=None, res_name_or_index=None,
                 sfntVersion="\000\001\000\000", flavor=None,
                 checkChecksums=False, verbose=None, recalcBBoxes=True,
                 allowVID=False, ignoreDecompileErrors=False,
                 recalcTimestamp=True, fontNumber=-1, lazy=None, quiet=None,
                 supressHints=False, showExtensionFlag=False):

        self.filePath = file
        self.supressHints = supressHints
        self.showExtensionFlag = showExtensionFlag
        TTFont. __init__(self, file, res_name_or_index=res_name_or_index,
                         sfntVersion=sfntVersion, flavor=flavor,
                         checkChecksums=checkChecksums, verbose=verbose,
                         recalcBBoxes=recalcBBoxes, allowVID=allowVID,
                         ignoreDecompileErrors=ignoreDecompileErrors,
                         recalcTimestamp=recalcTimestamp,
                         fontNumber=fontNumber, lazy=lazy, quiet=quiet)

    def _tableToXML(self, writer, tag, quiet=None, splitGlyphs=False):
        table = None
        if quiet is not None:
            from fontTools.misc.loggingTools import deprecateArgument
            deprecateArgument("quiet", "configure logging instead")
        if tag in self:
            table = self[tag]
            report = "Dumping '%s' table..." % tag
        else:
            report = "No '%s' table found." % tag
        log.info(report)
        if tag not in self:
            return
        xmlTag = tagToXML(tag)
        attrs = dict()
        if hasattr(table, "ERROR"):
            attrs['ERROR'] = "decompilation error"
        if table.__class__ == DefaultTable:
            attrs['raw'] = True
        writer.begintag(xmlTag, **attrs)
        writer.newline()
        if tag in ("glyf", "CFF "):
            dumpFont(writer, self.filePath, self.supressHints)
        elif tag in ("GSUB", "GPOS"):
            dumpOTLAsFeatureFile(writer, self, tag)
        else:
            table.toXML(writer, self)
        writer.endtag(xmlTag)
        writer.newline()
        writer.newline()


def shellcmd(cmdList):
    # In all cases, assume that cmdList does NOT specify the output file.
    # I use this because tx -dump -6 can be very large.
    tempPath = get_temp_file_path()
    cmdList.append(tempPath)
    subprocess.check_call(cmdList)
    with open(tempPath, "r", encoding="utf-8") as fp:
        data = fp.read()
    return data


def dumpFont(writer, fontPath, supressHints=False):
    dictTxt = shellcmd([TX_TOOL, "-dump", "-0", fontPath])
    if curSystem == "Windows":
        dictTxt = re.sub(r"[\r\n]+", "\n", dictTxt)
    dictTxt = re.sub(r"##[^\r\n]*Filename[^\r\n]+", "", dictTxt, 1).strip()
    dictLines = dictTxt.splitlines()
    writer.begintag("FontTopDict")
    writer.newline()
    for line in dictLines:
        writer._writeraw(line)
        writer.newline()
    writer.endtag("FontTopDict")
    writer.newline()

    if supressHints:
        charData = shellcmd([TX_TOOL, "-dump", "-6", "-n", fontPath])
    else:
        charData = shellcmd([TX_TOOL, "-dump", "-6", fontPath])

    if curSystem == "Windows":
        charData = re.sub(r"[\r\n]+", "\n", charData)
    charList = re.findall(r"[^ ]glyph\[[^]]+\] {([^,]+),[^\r\n]+,([^}]+)",
                          charData)
    if "cid.CIDFontName" in dictTxt:
        # fix glyph names to sort
        charList = [("cid%s" % (entry[0]).zfill(5),
                     entry[1]) for entry in charList]

    charList = sorted([entry[0] + entry[1] for entry in charList])
    charTxt = "\n".join(charList)
    writer.begintag("FontOutlines")
    writer.newline()
    for line in charTxt.splitlines():
        writer._writeraw(line)
        writer.newline()
    writer.endtag("FontOutlines")
    writer.newline()


@Timer(log, 'Done dumping TTX in %(time).3f seconds')
def ttnDump(input_file, output, options, showExtensionFlag, supressHints=False,
            supressVersions=False, supressTTFDiffs=False):
    log.info('Dumping "%s" to "%s"...', input_file, output)
    if options.unicodedata:
        from fontTools.unicode import setUnicodeData
        setUnicodeData(options.unicodedata)
    ttf = TTXNTTFont(input_file, 0, allowVID=options.allowVID,
                     ignoreDecompileErrors=options.ignoreDecompileErrors,
                     fontNumber=options.fontNumber, supressHints=supressHints,
                     showExtensionFlag=showExtensionFlag)

    kDoNotDumpTagList = ["GlyphOrder", "DSIG"]
    if options.onlyTables:
        onlyTables = sorted(options.onlyTables)
    else:
        onlyTables = sorted(ttf.keys())

    if options.skipTables:
        for tag in options.skipTables:
            if tag in onlyTables:
                onlyTables.remove(tag)

    for tag in kDoNotDumpTagList:
        if tag in onlyTables:
            onlyTables.remove(tag)

    # Zero values that always differ.
    if 'head' in onlyTables:
        head = ttf["head"]
        temp = head.created
        head.created = 0
        head.modified = 0
        head.magicNumber = 0
        head.checkSumAdjustment = 0
        if supressVersions:
            head.fontRevision = 0

    if 'hmtx' in onlyTables:
        hmtx = ttf["hmtx"]
        # hmtx must be decompiled *before* we zero
        # the hhea.numberOfHMetrics value
        temp = hmtx.metrics
        if supressTTFDiffs:
            try:
                del temp["CR"]
            except KeyError:
                pass
            try:
                del temp["NULL"]
            except KeyError:
                pass

    if 'hhea' in onlyTables:
        hhea = ttf["hhea"]
        temp = hhea.numberOfHMetrics
        hhea.numberOfHMetrics = 0

    if 'vmtx' in onlyTables:
        vmtx = ttf["vmtx"]
        # vmtx must be decompiled *before* we zero
        # the vhea.numberOfHMetrics value
        temp = vmtx.metrics
        if supressTTFDiffs:
            try:
                del temp["CR"]
            except KeyError:
                pass
            try:
                del temp["NULL"]
            except KeyError:
                pass

    if 'vhea' in onlyTables:
        vhea = ttf["vhea"]
        temp = vhea.numberOfVMetrics
        vhea.numberOfVMetrics = 0

    if supressVersions:
        if 'name' in onlyTables:
            name_table = ttf["name"]
            for namerecord in name_table.names:
                if namerecord.nameID == 3:
                    if namerecord.isUnicode():
                        namerecord.string = "VERSION SUPPRESSED".encode(
                            "utf-16be")
                    else:
                        namerecord.string = "VERSION SUPPRESSED"
                elif namerecord.nameID == 5:
                    if namerecord.platformID == 3:
                        namerecord.string = namerecord.string.split(';')[0]
                    else:
                        namerecord.string = "VERSION SUPPRESSED"

    if 'GDEF' in onlyTables:
        GDEF = ttf["GDEF"]
        gt = GDEF.table
        if gt.GlyphClassDef:
            gt.GlyphClassDef.Format = 0
        if gt.MarkAttachClassDef:
            gt.MarkAttachClassDef.Format = 0
        if gt.AttachList:
            if not gt.AttachList.Coverage.glyphs:
                gt.AttachList = None

        if gt.LigCaretList:
            if not gt.LigCaretList.Coverage.glyphs:
                gt.LigCaretList = None

    if 'cmap' in onlyTables:
        # remove mappings to notdef.
        cmapTable = ttf["cmap"]
        """ Force shared cmap tables to get separately decompiled.
        The _c_m_a_p.py logic will decompile a cmap from source data if an
        attempt is made to access a field which is not (yet) defined. Unicode
        (format 4) subtables are usually identical, and thus shared. When
        intially decompiled, the first gets fully decompiled, and then the
        second gets a reference to the 'cmap' dict of the first.
        When entries are removed from the cmap dict of the first subtable,
        that is the same as removing them from the cmap dict of the second.
        However, when later an attempt is made to reference the 'nGroups'
        field - which doesn't exist  in format 4 - the second table gets
        fully decompiled, and its cmap dict is rebuilt from the original data.
        """
        for cmapSubtable in cmapTable.tables:
            # if cmapSubtable.format != 4:
            #    continue
            delList = []
            for mapping in cmapSubtable.cmap.items():
                if mapping[1] == ".notdef":
                    delList.append(mapping)
                if supressTTFDiffs:
                    if mapping[1] == "CR":
                        delList.append(mapping)
                    if mapping[1] == "NULL":
                        delList.append(mapping)
            if delList:
                for charCode, _glyphName in delList:
                    try:
                        del cmapSubtable.cmap[charCode]
                    except KeyError:
                        pass

            if (cmapSubtable.format in [12, 14]) and hasattr(cmapSubtable,
                                                             "nGroups"):
                cmapSubtable.nGroups = 0
            if hasattr(cmapSubtable, "length"):
                cmapSubtable.length = 0

    if onlyTables:
        ttf.saveXML(output,
                    tables=onlyTables,
                    skipTables=options.skipTables,
                    splitTables=options.splitTables,
                    splitGlyphs=options.splitGlyphs,
                    disassembleInstructions=options.disassembleInstructions,
                    bitmapGlyphDataFormat=options.bitmapGlyphDataFormat)
    ttf.close()
    return ttf


def main(args=None):
    from fontTools import configLogger

    if args is None:
        args = sys.argv[1:]

    if ("-h" in args) or ("-u" in args):
        print(__help__)
        return 0

    if "-a" not in args:
        args.insert(0, "-a")  # allow virtual GIDS.

    if "-nh" in args:
        args.remove("-nh")
        supressHints = True
    else:
        supressHints = False

    if "-se" in args:
        args.remove("-se")
        showExtensionFlag = True
    else:
        showExtensionFlag = False

    if "-nv" in args:
        args.remove("-nv")
        supressVersions = True
    else:
        supressVersions = False

    if "-supressTTFDiffs" in args:
        args.remove("-supressTTFDiffs")
        supressTTFDiffs = True
    else:
        supressTTFDiffs = False

    try:
        jobs, options = ttx.parseOptions(args)
    except getopt.GetoptError as e:
        print("ERROR:", e, file=sys.stderr)
        return 2

    configLogger(level=options.logLevel)

    try:
        for action, input_file, output in jobs:
            if action != ttx.ttDump:
                log.error("ttxn can only dump font files.")
                return 1
            ttnDump(input_file, output, options, showExtensionFlag,
                    supressHints, supressVersions, supressTTFDiffs)
    except SystemExit:
        raise
    except TTLibError as e:
        log.error(e)
        return 1


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2017 Adobe. All rights reserved.

"""
Builds a CFF2 variable font from a designspace file and its UFO masters.
"""

import argparse
from ast import literal_eval
from copy import deepcopy
import logging
import os
import re
import sys

from fontTools import varLib
from fontTools.cffLib.specializer import commandsToProgram
from fontTools.designspaceLib import DesignSpaceDocument
from fontTools.misc.fixedTools import otRound
from fontTools.misc.psCharStrings import T2OutlineExtractor, T2CharString
from fontTools.ttLib import TTFont
from fontTools.varLib.cff import (CFF2CharStringMergePen,
                                  VarLibCFFPointTypeMergeError)

from afdko.fdkutils import validate_path

__version__ = '2.0.2'

STAT_FILENAME = 'override.STAT.ttx'


class CFF2VFError(Exception):
    """Base exception for buildcff2vf"""


# set up for printing progress notes
def progress(self, message, *args, **kws):
    # Note: message must contain the format specifiers for any strings in args.
    level = self.getEffectiveLevel()
    self._log(level, message, args, **kws)


PROGRESS_LEVEL = logging.INFO + 5
PROGESS_NAME = "progress"
logging.addLevelName(PROGRESS_LEVEL, PROGESS_NAME)
logger = logging.getLogger(__name__)
logging.Logger.progress = progress


def getSubset(subset_Path):
    with open(subset_Path, "rt") as fp:
        text_lines = fp.readlines()
    locationDict = {}
    cur_key_list = None
    for li, line in enumerate(text_lines):
        idx = line.find('#')
        if idx >= 0:
            line = line[:idx]
        line = line.strip()
        if not line:
            continue
        if line[0] == "(":
            cur_key_list = []
            location_list = literal_eval(line)
            for location_entry in location_list:
                cur_key_list.append(location_entry)
                if location_entry not in locationDict:
                    locationDict[location_entry] = []
        else:
            m = re.match(r"(\S+)", line)
            if m:
                if cur_key_list is None:
                    logger.error(
                        "Error parsing subset file. "
                        "Seeing a glyph name record before "
                        "seeing a location record.")
                    logger.error(f'Line number: {li}.')
                    logger.error(f'Line text: {line}.')
                for key in cur_key_list:
                    locationDict[key].append(m.group(1))
    return locationDict


def subset_masters(designspace, subsetDict):
    from fontTools import subset
    subset_options = subset.Options(notdef_outline=True, layout_features='*')
    for ds_source in designspace.sources:
        key = tuple(ds_source.location.items())
        included = set(subsetDict[key])
        ttf_font = ds_source.font
        subsetter = subset.Subsetter(options=subset_options)
        subsetter.populate(glyphs=included)
        subsetter.subset(ttf_font)
        subset_path = f'{os.path.splitext(ds_source.path)[0]}.subset.otf'
        logger.progress(f'Saving subset font {subset_path}')
        ttf_font.save(subset_path)
        ds_source.font = TTFont(subset_path)


class CompatibilityPen(CFF2CharStringMergePen):
    def __init__(self, default_commands,
                 glyphName, num_masters, master_idx, roundTolerance=0.5):
        super(CompatibilityPen, self).__init__(
            default_commands, glyphName, num_masters, master_idx,
            roundTolerance=0.5)
        self.fixed = False

    def add_point(self, point_type, pt_coords):
        if self.m_index == 0:
            self._commands.append([point_type, [pt_coords]])
        else:
            cmd = self._commands[self.pt_index]
            if cmd[0] != point_type:
                # Fix some issues that show up in some
                # CFF workflows, even when fonts are
                # topologically merge compatible.
                success, new_pt_coords = self.check_and_fix_flat_curve(
                    cmd, point_type, pt_coords)
                if success:
                    logger.progress(f"Converted between line and curve in "
                                    f"source font index '{self.m_index}' "
                                    f"glyph '{self.glyphName}', point index "
                                    f"'{self.pt_index}' at '{pt_coords}'. "
                                    f"Please check correction.")
                    pt_coords = new_pt_coords
                else:
                    success = self.check_and_fix_closepath(
                        cmd, point_type, pt_coords)
                    if success:
                        # We may have incremented self.pt_index
                        cmd = self._commands[self.pt_index]
                        if cmd[0] != point_type:
                            success = False
                if not success:
                    raise VarLibCFFPointTypeMergeError(
                        point_type, self.pt_index, self.m_index, cmd[0],
                        self.glyphName)
                self.fixed = True
            cmd[1].append(pt_coords)
        self.pt_index += 1

    def make_flat_curve(self, cur_coords):
        # Convert line coords to curve coords.
        dx = self.roundNumber(cur_coords[0] / 3.0)
        dy = self.roundNumber(cur_coords[1] / 3.0)
        new_coords = [dx, dy, dx, dy,
                      cur_coords[0] - 2 * dx,
                      cur_coords[1] - 2 * dy]
        return new_coords

    def make_curve_coords(self, coords, is_default):
        # Convert line coords to curve coords.
        if is_default:
            new_coords = []
            for cur_coords in coords:
                master_coords = self.make_flat_curve(cur_coords)
                new_coords.append(master_coords)
        else:
            cur_coords = coords
            new_coords = self.make_flat_curve(cur_coords)
        return new_coords

    def check_and_fix_flat_curve(self, cmd, point_type, pt_coords):
        success = False
        if (point_type == 'rlineto') and (cmd[0] == 'rrcurveto'):
            is_default = False  # the line is in the master font we are adding
            pt_coords = self.make_curve_coords(pt_coords, is_default)
            success = True
        elif (point_type == 'rrcurveto') and (cmd[0] == 'rlineto'):
            is_default = True  # the line is in the default font commands
            expanded_coords = self.make_curve_coords(cmd[1], is_default)
            cmd[1] = expanded_coords
            cmd[0] = point_type
            success = True
        return success, pt_coords

    def check_and_fix_closepath(self, cmd, point_type, pt_coords):
        """ Some workflows drop a lineto which closes a path.
        Also, if the last segment is a curve in one master,
        and a flat curve in another, the flat curve can get
        converted to a closing lineto, and then dropped.
        Test if:
        1) one master op is a moveto,
        2) the previous op for this master does not close the path
        3) in the other master the current op is not a moveto
        4) the current op in the otehr master closes the current path

        If the default font is missing the closing lineto, insert it,
        then proceed with merging the current op and pt_coords.

        If the current region is missing the closing lineto
        and therefore the current op is a moveto,
        then add closing coordinates to self._commands,
        and increment self.pt_index.

        Note that if this may insert a point in the default font list,
        so after using it, 'cmd' needs to be reset.

        return True if we can fix this issue.
        """
        if point_type == 'rmoveto':
            # If this is the case, we know that cmd[0] != 'rmoveto'

            # The previous op must not close the path for this region font.
            prev_moveto_coords = self._commands[self.prev_move_idx][1][-1]
            prv_coords = self._commands[self.pt_index - 1][1][-1]
            if prev_moveto_coords == prv_coords[-2:]:
                return False

            # The current op must close the path for the default font.
            prev_moveto_coords2 = self._commands[self.prev_move_idx][1][0]
            prv_coords = self._commands[self.pt_index][1][0]
            if prev_moveto_coords2 != prv_coords[-2:]:
                return False

            # Add the closing line coords for this region
            # so self._commands, then increment self.pt_index
            # so that the current region op will get merged
            # with the next default font moveto.
            if cmd[0] == 'rrcurveto':
                new_coords = self.make_curve_coords(prev_moveto_coords, False)
                cmd[1].append(new_coords)
            self.pt_index += 1
            return True

        if cmd[0] == 'rmoveto':
            # The previous op must not close the path for the default font.
            prev_moveto_coords = self._commands[self.prev_move_idx][1][0]
            prv_coords = self._commands[self.pt_index - 1][1][0]
            if prev_moveto_coords == prv_coords[-2:]:
                return False

            # The current op must close the path for this region font.
            prev_moveto_coords2 = self._commands[self.prev_move_idx][1][-1]
            if prev_moveto_coords2 != pt_coords[-2:]:
                return False

            # Insert the close path segment in the default font.
            # We omit the last coords from the previous moveto
            # is it will be supplied by the current region point.
            # after this function returns.
            new_cmd = [point_type, None]
            prev_move_coords = self._commands[self.prev_move_idx][1][:-1]
            # Note that we omit the last region's coord from prev_move_coords,
            # as that is from the current region, and we will add the
            # current pts' coords from the current region in its place.
            if point_type == 'rlineto':
                new_cmd[1] = prev_move_coords
            else:
                # We omit the last set of coords from the
                # previous moveto, as it will be supplied by the coords
                # for the current region pt.
                new_cmd[1] = self.make_curve_coords(prev_move_coords, True)
            self._commands.insert(self.pt_index, new_cmd)
            return True
        return False

    def getCharStrings(self, num_masters, private=None, globalSubrs=None,
                       default_idx=0):
        """ A command looks like:
        [op_name, [
            [source 0 arglist for op],
            [source 1 arglist for op],
            ...
            [source n arglist for op],
        I am not optimizing this there, as that will be done when
        the CFF2 Charstring is created in fontTools.varLib.build().

        If I did, I would have to rearrange the arguments to:
        [
        [arg 0 for source 0 ... arg 0 for source n]
        [arg 1 for source 0 ... arg 1 for source n]
        ...
        [arg M for source 0 ... arg M for source n]
        ]
        before calling specialize.
        """
        t2List = []
        merged_commands = self._commands
        for i in range(num_masters):
            commands = []
            for op in merged_commands:
                source_op = [op[0], op[1][i]]
                commands.append(source_op)
            program = commandsToProgram(commands)
            if self._width is not None:
                assert not self._CFF2, (
                    "CFF2 does not allow encoding glyph width in CharString.")
                program.insert(0, otRound(self._width))
            if not self._CFF2:
                program.append('endchar')
            charString = T2CharString(
                program=program, private=private, globalSubrs=globalSubrs)
            t2List.append(charString)
        # if default_idx is not 0, we need to move it to the right index.
        if default_idx:
            default_font_cs = t2List.pop(0)
            t2List.insert(default_idx, default_font_cs)
        return t2List


def _get_cs(charstrings, glyphName):
    if glyphName not in charstrings:
        return None
    return charstrings[glyphName]


def do_compatibility(vf, master_fonts, default_idx):
    default_font = vf
    default_charStrings = default_font['CFF '].cff.topDictIndex[0].CharStrings
    glyphOrder = default_font.getGlyphOrder()
    charStrings = [
        font['CFF '].cff.topDictIndex[0].CharStrings for font in master_fonts]

    for gname in glyphOrder:
        all_cs = [_get_cs(cs, gname) for cs in charStrings]
        if len([gs for gs in all_cs if gs is not None]) < 2:
            continue
        # remove the None's from the list.
        cs_list = [cs for cs in all_cs if cs]
        num_masters = len(cs_list)
        default_charstring = default_charStrings[gname]
        compat_pen = CompatibilityPen([], gname, num_masters, 0)
        default_charstring.outlineExtractor = T2OutlineExtractor
        default_charstring.draw(compat_pen)

        # Add the coordinates from all the other regions to the
        # blend lists in the CFF2 charstring.
        region_cs = cs_list[:]
        del region_cs[default_idx]
        for region_idx, region_charstring in enumerate(region_cs, start=1):
            compat_pen.restart(region_idx)
            region_charstring.draw(compat_pen)
        if compat_pen.fixed:
            fixed_cs_list = compat_pen.getCharStrings(
                num_masters, private=default_charstring.private,
                globalSubrs=default_charstring.globalSubrs,
                default_idx=default_idx)
            cs_list = list(cs_list)
            for i, cs in enumerate(cs_list):
                mi = all_cs.index(cs)
                charStrings[mi][gname] = fixed_cs_list[i]


def otfFinder(s):
    return s.replace('.ufo', '.otf')


def suppress_glyph_names(tt_font):
    postTable = tt_font['post']
    postTable.formatType = 3.0
    postTable.compile(tt_font)


def remove_mac_names(tt_font):
    name_tb = tt_font['name']
    name_tb.names = [nr for nr in name_tb.names if nr.platformID != 1]


def update_stat_name_ids(tt_font):
    """
    The STAT spec says that axes must point to the same name ID used
    in the fvar so check here and update if they are different.
    """
    fvar = tt_font['fvar']
    stat = tt_font['STAT']
    fvar_axis_names = {}

    for axis in fvar.axes:
        fvar_axis_names[axis.axisTag] = axis.axisNameID

    for axis in stat.table.DesignAxisRecord.Axis:
        fvar_id = fvar_axis_names.get(axis.AxisTag)
        if fvar_id is None:
            # Not required for all STAT axes to be in fvar
            continue
        if axis.AxisNameID != fvar_id:
            axis.AxisNameID = fvar_id


def validate_stat_axes(tt_font):
    """
    Ensure all axes defined in fvar also exist in the STAT table
    """
    fvar = tt_font['fvar']
    stat = tt_font['STAT']
    fvar_axis_tags = [axis.axisTag for axis in fvar.axes]
    stat_axis_tags = [axis.AxisTag for axis in
                      stat.table.DesignAxisRecord.Axis]
    diff = set(fvar_axis_tags) - set(stat_axis_tags)
    if diff:
        raise CFF2VFError(
            f'All fvar axes must also be defined in the STAT table. '
            f'Axes for {str(list(diff))} are missing.'
        )


def validate_stat_values(ttFont):
    """
    Check axis values in the STAT table to ensure they are within the ranges
    defined in the fvar
    """
    fvar = ttFont['fvar']
    stat = ttFont['STAT']

    logger.progress('Validating STAT axis values...')
    errors = []
    stat_range_vals = {}

    if hasattr(stat.table.AxisValueArray, "AxisValue"):
        for av in stat.table.AxisValueArray.AxisValue:
            axis_tag = stat.table.DesignAxisRecord.Axis[av.AxisIndex].AxisTag
            if axis_tag not in stat_range_vals:
                stat_range_vals[axis_tag] = []
            if hasattr(av, 'NominalValue'):
                stat_range_vals[axis_tag].append(av.NominalValue)
                if not av.RangeMinValue <= av.NominalValue <= av.RangeMaxValue:
                    errors.append(
                        f'Invalid default value {av.NominalValue} for range '
                        f'{av.RangeMinValue} - {av.RangeMaxValue}'
                    )
            if hasattr(av, 'RangeMaxValue'):
                stat_range_vals[axis_tag].append(av.RangeMaxValue)
            if hasattr(av, 'RangeMinValue'):
                stat_range_vals[axis_tag].append(av.RangeMinValue)
            if hasattr(av, 'Value'):
                stat_range_vals[axis_tag].append(av.Value)

    for axis in fvar.axes:
        stat_ref = stat_range_vals.get(axis.axisTag)
        if stat_ref is None:
            continue
        out_of_range = []
        for val in stat_ref:
            if (val > axis.maxValue and int(val) != 32767) or \
                    (val < axis.minValue and int(val) != -32767):
                out_of_range.append(val)
        if out_of_range:
            expected_range = f'{axis.minValue} - {axis.maxValue}'
            errors.append(
                f'{axis.axisTag} values {str(sorted(set(out_of_range)))} are '
                f'outside of range {expected_range} specified in fvar'
            )
    if errors:
        msg = '\n'.join(errors)
        raise CFF2VFError(f'Invalid STAT table. {msg}')


def import_stat_override(tt_font, stat_file_path):
    if 'STAT' in tt_font:
        logger.warning(
            f'Overwriting existing STAT table with {stat_file_path}.'
        )
    tt_font.importXML(stat_file_path)


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for debug mode'
    )
    parser.add_argument(
        '-d',
        '--designspace',
        metavar='PATH',
        dest='design_space_path',
        type=validate_path,
        help='path to design space file',
        required=True
    )
    parser.add_argument(
        '-o',
        '--output',
        dest='var_font_path',
        metavar='PATH',
        help='path to output variable font file. Default is base name\n'
        'of the design space file.',
    )
    parser.add_argument(
        '-k',
        '--keep-glyph-names',
        action='store_true',
        help='Preserve glyph names in output variable font\n'
        "(using 'post' table format 2).",
    )
    parser.add_argument(
        '--omit-mac-names',
        action='store_true',
        help="Omit Macintosh strings from 'name' table.",
    )
    parser.add_argument(
        '-c',
        '--check-compat',
        dest='check_compatibility',
        action='store_true',
        help='Check outline compatibility in source fonts, and fix flat\n'
        'curves.',
    )
    parser.add_argument(
        '-i',
        '--include-glyphs',
        dest='include_glyphs_path',
        metavar='PATH',
        type=validate_path,
        help='Path to file containing a python dict specifying which\n'
        'glyph names should be included from which source fonts.'
    )
    options = parser.parse_args(args)

    if not options.var_font_path:
        var_font_path = f'{os.path.splitext(options.design_space_path)[0]}.otf'
        options.var_font_path = var_font_path

    if not options.verbose:
        level = PROGRESS_LEVEL
        logging.basicConfig(level=level, format="%(message)s")
    else:
        level = logging.INFO
        logging.basicConfig(level=level)
    logger.setLevel(level)

    return options


def main(args=None):
    options = get_options(args)

    if os.path.exists(options.var_font_path):
        os.remove(options.var_font_path)

    designspace = DesignSpaceDocument.fromfile(options.design_space_path)
    ds_data = varLib.load_designspace(designspace)
    master_fonts = varLib.load_masters(designspace, otfFinder)
    logger.progress("Reading source fonts...")
    for i, master_font in enumerate(master_fonts):
        designspace.sources[i].font = master_font

    # Subset source fonts
    if options.include_glyphs_path:
        logger.progress("Subsetting source fonts...")
        subsetDict = getSubset(options.include_glyphs_path)
        subset_masters(designspace, subsetDict)

    if options.check_compatibility:
        logger.progress("Checking outline compatibility in source fonts...")
        font_list = [src.font for src in designspace.sources]
        default_font = designspace.sources[ds_data.base_idx].font
        vf = deepcopy(default_font)
        # We copy vf from default_font, because we use VF to hold
        # merged arguments from each source font charstring - this alters
        # the font, which we don't want to do to the default font.
        do_compatibility(vf, font_list, ds_data.base_idx)

    logger.progress("Building variable OTF (CFF2) font...")
    # Note that we now pass in the design space object, rather than a path to
    # the design space file, in order to pass in the modified source fonts
    # fonts without having to recompile and save them.
    try:
        varFont, _, _ = varLib.build(designspace, otfFinder)
    except VarLibCFFPointTypeMergeError:
        logger.error("The input set requires compatibilization. Please try "
                     "again with the -c (--check-compat) option.")
        return 0

    if not options.keep_glyph_names:
        suppress_glyph_names(varFont)

    if options.omit_mac_names:
        remove_mac_names(varFont)

    stat_file_path = os.path.join(
        os.path.dirname(options.var_font_path), STAT_FILENAME)
    if os.path.exists(stat_file_path):
        logger.progress("Importing STAT table override...")
        import_stat_override(varFont, stat_file_path)

    validate_stat_axes(varFont)
    validate_stat_values(varFont)
    update_stat_name_ids(varFont)

    varFont.save(options.var_font_path)
    logger.progress(f"Built variable font '{options.var_font_path}'")


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2014 Adobe. All rights reserved.

"""
proofpdf.py. A wrapper for the fontpdf.py module. This script verifies
the existence of the specified font files, creates a font class object
with the call-backs required by fontpdf, and translates the command line
options to arguments for the fontpdf module; the latter produces a proof
file using the provided options annd font instance.
"""

import copy
import os
import platform
import re
import sys
import tempfile
import time
import traceback

from fontTools.ttLib import TTFont, getTableModule, TTLibError

from afdko import fdkutils
from afdko.pdflib import ttfpdf, otfpdf
from afdko.pdflib.fontpdf import (
	FontPDFParams, makePDF, makeFontSetPDF, kDrawTag, kDrawPointTag,
	kShowMetaTag, params_doc,
	)

curSystem = platform.system()


__usage__ = """
proofpdf v1.21 Aug 28 2018
proofpdf [-h] [-u]
proofpdf -help_params
proofpdf [-g <glyph list>] [-gf <filename>] [-gpp <number>] [-pt <number>] [-dno] [-baseline <number>] [-black] [-lf <filename>] [-select_hints <0,1,2..> ]  \
[-o <PDF file path> ] -hintplot ] [-charplot] [-digiplot] [-fontplot]  [-waterfallplot] [-wfr <point size list>] font-path1  fontpath2 ...

Glyph Proofing program for OpenType fonts
"""
__help__= __usage__ + """

"charplot", "digiplot", "fontplot", "hintplot", and "waterfallplot" are
all command files that call the proofpdf script with different options.
proofpdf takes as options a list of fonts, and an optional list of
glyphs, and prints a PDF file for the specified font, showing the glyphs
as specified by options.

The five main options, "-charplot", "-digiplot", "-fontplot","-hintplot"
and "waterfallplot", each set a bunch of lower level parameters in order
to produce a particular page layout. All these low-level parameters can
be set by command-line options.

Options:

-u Shows usage

-h Print help

-help_params Shows help abut the low level parameters.

-q Quiet mode

-charplot Sets parameters to show the glyph outline with most labels;
the point labels show only the point position. No hints or alignment
zones are shown; all glyph metric info is shown. Default is one glyph
per page.

-digiplot Sets parameters to show 2 copies of each glyph. The first is
shown with the em-box and the meta data block is filled with text from
an error log file, if one exists. The second shows the outline with some
labels; the point labels show only the point type. Default is one glyph
per page.

-repeatIndex <index>. Used in conjunction with the digiplot command. By default,
options after the 'digiplot' option will affect only the left-side copy of the
glyph. The 'repeatIndex' option can be used to specify that another copy of in
the set  of repeated glyph blocks will be affected by subsequent arguments.
Use '-repeatIndex 0" to specify the left-side glyph block, and '-repeatIndex 1'
to specify the right-side glyph glyph block, as the target for following
arguments.


-fontplot Sets parameters to show the filled glyph outline with no
labels, and some glyph metrics. Default is 156 glyphs per page.

-fontplot2 Sets parameters to show the filled glyph outline with no
labels, and a box showing em-box and height and width, with center
registration marks. Default is 156 glyphs per page.

-fontsetplot. Sets parameters to compare glyphs between different fonts. This
options figures out how many glyphs will fit across a page using a fixed spacing
of em size. It then divides the list of glyphs into groups of of this size. for
each group, the glyphs are shown in  a single line with fixed spacing for each
font in turn. This plot is useful for quickly seeing that the glyphs have the
same GID order between fonts, and that the glyph shape is correct by visually
comparing it to the nominally equivalent glyphs in the same column. Glyphs are
shown in glyph ID order. The font list is shown in a sorted order. The sorting
is first by the length of the charset, then by the charset lists, then by the
font PS name.

-alpha   Sorts the glyphs in alphabetic order by glyph name. Useful when
comparing proofs of fonts with the same charset, but different glyph order.

-hintplot Sets parameters to show the glyph outline with no labels, with
all hints and alignment zones, and with just the name and BBox info.
Default is one glyph per page.

-waterfallplot Sets parameters to show all the glyphs in a font in a series  of
waterfalls of decreasing point size. This options figures out how many glyphs
will fit across a page at the maximum point size, using the average character
width. It then divides the list of glyphs into groups of of this size, and shows
a waterfall for each group. The default list of point sizes for the waterfall
is: (36,24,20,18,16,14,12,10,9,8,7,6)

Note that this option  works very differently form the other proofs, in that the
font data is embedded, and the glyphs are requested by character code. In all
the other modes, the glyph outlines are drawn and filled with PDF drawing
operators, and the font is not embedded. This because the purpose is to check
hint quality at different point sizes, and this can be done only when the font
data is embedded and the glyphs are rasterized by the system or application.

Warning: this option does not yet work with Truetype or CID-keyed fonts.

-wfr <point size list> Overrides the default list of point sizes for the
waterfall mode. Has no effect if '-waterfallplot' is not specified. Example:
   -wfr 72,48,36,24,12,8
Note that the point sizes are separated by commas, with no white space allowed
in the list.


-g <glyphID1>,<glyphID2>,...,<glyphIDn>   Show only the specified list
of glyphs. The list must be comma-delimited, with no white space. The
glyph ID's may be glyph indices, glyph names, or glyph CID's. If the latter,
the CID value must be prefixed with the string "cid". There must be no
white-space in the glyph list. Examples: proofpdf -g A,B,C,69 myFont
proofpdf -g cid01030,cid00034,cid03455,69 myCIDFont.otf

A range of glyphs may be specified by providing two names separated only
by a hyphen:
fontplot -g zero-nine,onehalf myFont
Note that the range will be resolved by filling in the list of glyph
indices (GI) between the GI's of the two glyphs of the range, not by
alphabetic name order.

-gf <file name>   Hint only the list of glyphs contained in the
specified file, The file must contain a comma-delimited list of glyph
identifiers. Any number of space, tab, and new-line characters are
permitted between glyphnames and commas.

-gpp <number> Set the number of glyphs per page. The glyphs are scaled
so that this many glyphs will fit on a page.

-pt <number> Set point size for the glyphs. If supplied, -gpp will be
ignored.

-o <file path> Path name for output pdf file. Must end in "pdf". If
omitted, the default file name is <font name> + ".pdf"

-dno  Do not open the PDF file when done; default is to do so, except for
digiplot. Note: On Mac OS, if the  a PDF reader is already open and showing the
file, you will need to close the file window and re-open in order to see the new
contents. Also, The script does not choose the program to open the PDF; that is
handled by the computer operating system.

-do  Do  open the PDF file when done. Useful for digiplot, where the default is
to not open the pdf file.

-baseline Set baseline font font. This is is otherwise set
to the vale in the parent font BASE table. If the BASE table
is not present, it is set to -120 for CID keyed fonts, else
0.

-black Will set all colors to black; useful when planning to print the PDF.

-lf <filename> Path name to CID layout file. If supplied, and proofing a
CID-keyed font, glyphs will be shown in layout file order, and the
hint dir and row font dir name will be shown in the descriptive meta-data.

-select_hints <list of int replacement block indices> When showing
hints, will show only the specified hint replacement blocks. Example:
      -select_hints 3  will show only the 4th hint replacement block
      -select_hints 0,3  will show only the first and 4th hint replacement block.
Note that the index list must be comma-delimited with no white-space.

-v	Show the vertical metrics, but not horizontal metrics.

-vh Show both the vertical and horizontal metrics.

--<name> <value> If prepended with two hyphens, then this is interpreted
as a name of a low level parameter, which is set to the following value.
Note that any such options should follow the use of any of the five main
parameters, as -hintplot, -charplot, and -digiplot set a number of low
level parameters. Use "-help_params" to see the documentation for the
low level parameters. All sizes are expressed in points, where 72 points
= 1 inch. Fonts must be PostScript.
Examples:
--pageTitleFont "MinionPro-Bold"  #Change the page title font.
--pointLabelFont "Times-Bold"  #Change the font for point labels
--pointLabelColorRGB "(1.0, 0, 0)" #Change the color of the point labels
to red. The value is a triplet of (R G, B) values, where 0 is dark and 1.0 is
light. Note that the RGB values must be enclosed in quotes and
parentheses, and commas separated, as shown. "(0,0,0)" is black,
"(1.0, 1.0, 1.0)" is page white, "(1.0, 0, 0)" is red.
--pointLabelSize 12  #Change the size of the point label text.
"""


def logMsg(*args):
	for arg in args:
		print(arg)

class FDKEnvironmentError(Exception):
	pass

class OptionParseError(Exception):
	pass

class FontError(Exception):
	pass

def CheckEnvironment():
	tx_path = 'tx'
	txError = 0

	command = "%s -u 2>&1" % tx_path
	report = fdkutils.runShellCmd(command)
	if "options" not in report:
		txError = 1

	if txError:
		logMsg("Please check PATH and/or re-install the afdko. Cannot find: "
									"< %s >." % (tx_path))
		logMsg("or the files referenced by the shell script is missing.")
		raise FDKEnvironmentError

	return tx_path


def fixGlyphNames(glyphName):
	glyphRange = glyphName.split("-")
	if len(glyphRange) >1:
		g1 = fixGlyphNames(glyphRange[0])
		g2 =  fixGlyphNames(glyphRange[1])
		glyphName =  "%s-%s" % (g1, g2)

	elif glyphName[0] == "/":
		glyphName = "cid" + glyphName[1:].zfill(5)

	elif glyphName.startswith("cid") and (len(glyphName) < 8):
		return "cid" + glyphName[3:].zfill(5)

	return glyphName

def parseGlyphListArg(glyphString):
	glyphString = re.sub(r"[ \t\r\n,]+",  ",",  glyphString)
	glyphList = glyphString.split(",")
	glyphList = map(fixGlyphNames, glyphList)
	glyphList =  filter(None, glyphList)
	return glyphList

def expandRanges(tag):
	ptRange = tag.split("-")
	if len(ptRange) > 1:
		g1 = int(ptRange[0])
		g2 =  int(ptRange[1])
		ptRange = range(g1, g2+1)
	else:
		ptRange = [int(tag)]
	return ptRange

def parsePtSizeListArg(ptSizeString):
	ptSizeString = re.sub(r"[ \t\r\n,]+",  ",",  ptSizeString)
	ptsizeList = ptSizeString.split(",")
	ptEntryList = map(expandRanges, ptsizeList)
	ptSizeList = []
	for ptEntry in ptEntryList:
		ptSizeList.extend(ptEntry)
	return ptSizeList


def parseLayoutFile(filePath):
	layoutDict = None

	try:
		fp = open(filePath, "rt")
		data = fp.read()
		fp.close()
	except (IOError,OSError):
		raise OptionParseError("Option Error: could not open and read the layout file <%s>." %  filePath)

	entryList = re.findall(r"(\d+)\s+(\S+)\s+(\S+)\s+(\S+)", data)
	if len(entryList) < 2:
		print("Error: Failed to parse layout file %s. Did not match expected entry format." % filePath)
		raise OptionParseError
	print("Found %s entries in layout file %s. Mac CID: %s" % (len(entryList), os.path.basename(filePath), entryList[-1][0]))
	layoutDict = {}
	index = 0
	for entry in entryList:
		gname = "cid" + entry[0].zfill(5)
		layoutDict[gname] = [entry[1], entry[2], entry[3]] # add cide key-> family, face dir, name
		reversekey = repr(entry[1:])
		layoutDict[reversekey] = [gname, entry[1], entry[2], entry[3]] # add name key -> family, face, cid as cidXXXX
		entryList = None
	if "cid00000" in layoutDict:
		layoutDict[".notdef"] = layoutDict["cid00000"]
	return layoutDict


def setDefaultDigiplotRightSideOptions(rightGlyphParams):
	exec("rightGlyphParams." + kDrawTag + "EMBox = 0")
	exec("rightGlyphParams." + kDrawTag + "XAdvance = 0")
	exec("rightGlyphParams." + kDrawTag + "ContourLabels = 1")
	exec("rightGlyphParams." + kDrawTag + "Outline = 1")
	exec("rightGlyphParams." + kDrawTag + "CenterMarks = 0")
	exec("rightGlyphParams." + kDrawTag + "GlyphBox = 0")
	exec("rightGlyphParams." + kDrawPointTag + "PointMarks = 1")
	exec("rightGlyphParams." + kDrawPointTag + "BCPMarks = 1")
	exec("rightGlyphParams." + kDrawPointTag + "PointLabels = 1")
	rightGlyphParams.pointLabel_doPointType  = 1 # add point type to point label .
	rightGlyphParams.pointLabel_doPosition  = 0 # add point position to point label .
	rightGlyphParams.pointLabel_doPointIndex  = 0 # add index to point label .
	rightGlyphParams.metaDataAboveGlyph = 0 # write meta data below glyph square.
	if rightGlyphParams.rt_optionLayoutDict:
		exec("rightGlyphParams." + kShowMetaTag + "WidthOnly = 1")
		exec("rightGlyphParams." + kShowMetaTag + "HintDir = 1")
		exec("rightGlyphParams." + kShowMetaTag + "RowFont = 1")
		exec("rightGlyphParams." + kShowMetaTag + "Hints = 0")
		exec("rightGlyphParams." + kShowMetaTag + "SideBearings = 0")
		exec("rightGlyphParams." + kShowMetaTag + "V_SideBearings = 0")

def getOptions(params):
	i = 1
	numOptions = len(sys.argv)
	makeRepeatParams = 0
	params.quietMode = False
	while i < numOptions:
		arg = sys.argv[i]
		if params.rt_fileList and arg[0] == "-":
			raise OptionParseError("Option Error: All file names must follow all other params <%s>." % arg)

		if arg == "-h":
			print(__help__)
			sys.exit()

		elif arg == "-u":
			print(__usage__)
			sys.exit()

		elif arg == "-help_params":
			print(params_doc)
			sys.exit()

		elif arg[:2] == "--":
			name = arg[2:]
			i +=1
			if not hasattr(params, name):
				print("Error: low level parameter name %s does not exist in list of display parameters." % name)
				raise OptionParseError
			else:
				value = None
				arg = sys.argv[i]
				if name.endswith("RGB"):
					match = re.search(r"\(\s*([0-9.-]+)\s*,\s*([0-9.-]+)\s*,\s*([0-9.-]+)\s*\)", arg)
					if match:
						value = ( eval(match.group(1)), eval(match.group(2)), eval(match.group(3)) )
					else:
						print("Error: parameter %s must take an RGB triplet, enclosed in quotes, such as '( 1.0, 0, 0)' for red." % name)
						raise OptionParseError
				elif name == "pageSize":
					match = re.search(r"\(\s*([0-9.-]+)\s*,\s*([0-9.-]+)\s*\)", arg)
					if match:
						value = ( eval(match.group(1)), eval(match.group(2)) )
					else:
						print("Error: parameter %s must take an (x,y) pair, enclosed in quotes, such as ' (612.0, 792.0)'." % name)
						raise OptionParseError
				elif name.endswith("Font") or name.endswith("Path"):
					value = arg # it is text - don't need to convert
				else:
					# Must be numeric.
					try:
						value = int(arg)
					except ValueError:
						print("Error: value for parameter %s must be a number." % name)
						raise OptionParseError
				if value != None:
					exec("params.%s = value" % (name))

		elif arg == "-hintplot":
			exec("params." + kDrawTag + "EMBox = 1")
			exec("params." + kDrawTag + "Baseline = 0")
			exec("params." + kDrawTag + "ContourLabels = 0")
			exec("params." + kDrawPointTag + "PointLabels = 0")
			exec("params." + kShowMetaTag + "Outline = 0")
			exec("params." + kShowMetaTag + "Parts = 0")
			exec("params." + kShowMetaTag + "Paths = 0")
			exec("params." + kShowMetaTag + "Name = 1")
			exec("params." + kShowMetaTag + "BBox = 0")
			exec("params." + kShowMetaTag + "SideBearings = 1")
			exec("params." + kShowMetaTag + "V_SideBearings = 0")
			exec("params." + kShowMetaTag + "Parts = 0")
			exec("params." + kShowMetaTag + "Paths = 0")
			params.glyphsPerPage = 4
			params.pointLabelSize = 32
		elif arg == "-charplot":
			exec("params." + kDrawTag + "HHints = 0")
			exec("params." + kDrawTag + "VHints = 0")
			exec("params." + kDrawTag + "BlueZones = 0")
			exec("params." + kDrawTag + "Baseline = 0")
			exec("params." + kDrawTag + "EMBox = 0")
			params.pointLabel_doPointType  = 0 # add point type to point label .
			params.pointLabel_doPosition  = 1 # add point position to point label .
			params.pointLabel_doPointIndex  = 0 # add index to point label .
		elif arg == "-digiplot":
			params.rt_repeats = 2
			makeRepeatParams = 1
			exec("params." + kDrawTag + "HHints = 0")
			exec("params." + kDrawTag + "VHints = 0")
			exec("params." + kDrawTag + "BlueZones = 0")
			exec("params." + kDrawTag + "Baseline = 0")
			exec("params." + kDrawTag + "EMBox = 1")
			exec("params." + kDrawTag + "XAdvance = 1")
			exec("params." + kDrawTag + "CenterMarks = 1")
			exec("params." + kDrawTag + "ContourLabels = 0")
			exec("params." + kDrawTag + "Outline = 1")
			exec("params." + kDrawPointTag + "PointMarks = 0")
			exec("params." + kDrawPointTag + "BCPMarks = 0")
			exec("params." + kDrawPointTag + "PointLabels = 0")
			params.openPDFWhenDone = 0
			params.glyphsPerPage = 2
			params.descenderSpace = -120
			params.userBaseLine = -120
			# params.errorLogFilePath = "Error_file.log"
			params.errorLogColumnHeight = 250
			params.metaDataAboveGlyph = 0 # write meta data below glyph square.
			params.userPtSize = 255
			params.pageTopMargin  = 0
			params.pageBottomMargin = 0
			params.pageLeftMargin = 20
			params.pageRightMargin  = 20
			params.glyphVPadding  = 0
			params.glyphHPadding   = 0

			rightGlyphParams = copy.copy(params)
			setDefaultDigiplotRightSideOptions(rightGlyphParams)
			params.rt_repeatParamList = [params, rightGlyphParams]

		elif arg == "-repeatIndex":
			index = sys.argv[i+1]
			try:
				index = eval(index)
				if (index < 0) or (index >= len(params.rt_repeatParamList)):
					raise IndexError
			except:
				logMsg( "\tError. Option '%s' must be followed by an index number value, 0 or 1." % (arg))
			params = params.rt_repeatParamList[index]


		elif arg == "-fontplot":
			exec("params." + kDrawTag + "HHints = 0")
			exec("params." + kDrawTag + "VHints = 0")
			exec("params." + kDrawTag + "BlueZones = 0")
			exec("params." + kDrawTag + "Baseline = 0")
			exec("params." + kDrawTag + "EMBox = 1")
			exec("params." + kDrawTag + "XAdvance = 1")
			exec("params." + kDrawTag + "ContourLabels = 0")
			exec("params." + kDrawTag + "Outline = 1")
			exec("params." + kDrawPointTag + "PointMarks = 0")
			exec("params." + kDrawPointTag + "BCPMarks = 0")
			exec("params." + kDrawPointTag + "PointLabels = 0")
			exec("params." + kShowMetaTag + "Outline = 0") # shows filled outline in  lower right of meta data area under glyph outline.
			exec("params." + kShowMetaTag + "Name = 1")
			exec("params." + kShowMetaTag + "BBox = 0")
			exec("params." + kShowMetaTag + "SideBearings = 1")
			exec("params." + kShowMetaTag + "V_SideBearings = 0")
			exec("params." + kShowMetaTag + "Parts = 0")
			exec("params." + kShowMetaTag + "Paths = 0")
			params.DrawFilledOutline = 1
			params.metaDataXoffset = 0
			params.metaDataTextSize = 80
			params.metaDataNameSize = 80
			params.glyphsPerPage = 156
		elif arg == "-fontplot2":
			exec("params." + kDrawTag + "HHints = 0")
			exec("params." + kDrawTag + "VHints = 0")
			exec("params." + kDrawTag + "BlueZones = 0")
			exec("params." + kDrawTag + "Baseline = 0")
			exec("params." + kDrawTag + "EMBox = 0")
			exec("params." + kDrawTag + "XAdvance = 0")
			exec("params." + kDrawTag + "ContourLabels = 0")
			exec("params." + kDrawTag + "Outline = 1")
			exec("params." + kDrawTag + "CenterMarksWithBox = 1")
			exec("params." + kDrawPointTag + "PointMarks = 0")
			exec("params." + kDrawPointTag + "BCPMarks = 0")
			exec("params." + kDrawPointTag + "PointLabels = 0")
			exec("params." + kShowMetaTag + "Outline = 0") # shows filled outline in  lower right of meta data area under glyph outline.
			exec("params." + kShowMetaTag + "Name = 1")
			exec("params." + kShowMetaTag + "BBox = 0")
			exec("params." + kShowMetaTag + "SideBearings = 1")
			exec("params." + kShowMetaTag + "V_SideBearings = 0")
			exec("params." + kShowMetaTag + "Parts = 0")
			exec("params." + kShowMetaTag + "Paths = 0")
			exec("params." + kShowMetaTag + "Hints = 0")
			params.DrawFilledOutline = 1
			params.metaDataXoffset = 0
			params.metaDataTextSize = 80
			params.metaDataNameSize = 80
			params.glyphsPerPage = 156
			params.metaDataAboveGlyph = 0
		elif arg == "-fontsetplot":
			exec("params." + kDrawTag + "HHints = 0")
			exec("params." + kDrawTag + "VHints = 0")
			exec("params." + kShowMetaTag + "Hints = 0")
			exec("params." + kDrawTag + "BlueZones = 0")
			exec("params." + kDrawTag + "Baseline = 0")
			exec("params." + kDrawTag + "EMBox = 0")
			exec("params." + kDrawTag + "XAdvance = 0")
			exec("params." + kDrawTag + "ContourLabels = 0")
			exec("params." + kDrawTag + "Outline = 1")
			exec("params." + kDrawPointTag + "PointMarks = 0")
			exec("params." + kDrawPointTag + "BCPMarks = 0")
			exec("params." + kDrawPointTag + "PointLabels = 0")
			exec("params." + kShowMetaTag + "Outline = 0") # shows filled outline in  lower right of meta data area under glyph outline.
			exec("params." + kShowMetaTag + "Name = 0")
			exec("params." + kShowMetaTag + "BBox = 0")
			exec("params." + kShowMetaTag + "SideBearings = 0")
			exec("params." + kShowMetaTag + "V_SideBearings = 0")
			exec("params." + kShowMetaTag + "Parts = 0")
			exec("params." + kShowMetaTag + "Paths = 0")
			exec("params." + kShowMetaTag + "Hints = 0")
			params.DrawFilledOutline = 1
			params.rt_doFontSet = 1
			if params.userPtSize is None: # set default point size.
				params.userPtSize = 12
		elif arg == "-alpha":
			params.rt_alphaSort = 1
		elif arg == "-waterfallplot":
			if not params.waterfallRange:
				params.waterfallRange = (36,24,20,18,16,14,12,10,9,8,7,6)
		elif arg == "-wfr":
			i = i +1
			sizeString = sys.argv[i]
			if sizeString[0] == "-":
				raise OptionParseError("Option Error: it looks like the first item in the waterfall size list following '-wfr' is another option.")
			params.waterfallRange = parsePtSizeListArg(sizeString)
		elif arg == "-gpp":
			i = i +1
			try:
				params.glyphsPerPage = int( sys.argv[i])
			except ValueError:
				raise OptionParseError("Option Error: -glyphsPerPage must be followed by an integer number")
		elif arg == "-pt":
			i = i +1
			try:
				params.userPtSize = int( sys.argv[i])
			except ValueError:
				raise OptionParseError("Option Error: -pt must be followed by an integer number")
		elif arg == "-g":
			i = i +1
			glyphString = sys.argv[i]
			if glyphString[0] == "-":
				raise OptionParseError("Option Error: it looks like the first item in the glyph list following '-g' is another option.")
			params.rt_optionGlyphList += parseGlyphListArg(glyphString)
		elif arg == "-gf":
			i = i +1
			rt_filePath = sys.argv[i]
			if rt_filePath[0] == "-":
				raise OptionParseError("Option Error: it looks like the the glyph list file following '-gf' is another option.")
			try:
				gf = open(rt_filePath, "rt")
				glyphString = gf.read()
				gf.close()
			except (IOError,OSError):
				raise OptionParseError("Option Error: could not open glyph list file <%s>." %  rt_filePath)
			params.rt_optionGlyphList += parseGlyphListArg(glyphString)
		elif arg == "-o":
			i = i +1
			params.rt_pdfFileName = sys.argv[i]
		elif arg == "-dno":
			params.openPDFWhenDone = 0
		elif arg == "-do":
			params.openPDFWhenDone = 1
		elif arg == "-q":
			params.quietMode = True
		elif arg == "-black":
			params.emboxColorRGB = params.baselineColorRGB = params.xAdvanceColorRGB = params.contourLabelColorRGB = params.pointLabelColorRGB = params.MetaDataGlyphColorRGB = (0,0,0)
			params.hintColorRGB = params.hintColorOverlapRGB = params.alignmentZoneColorRGB = (0.5,0.5,0.5)
		elif arg == "-lf":
			i = i +1
			rt_filePath = sys.argv[i]
			if rt_filePath[0] == "-":
				raise OptionParseError("Option Error: it looks like the the layout file following '-gf' is another option.")
			if not os.path.exists(rt_filePath):
				raise OptionParseError("Option Error: The layout file %s does not exist." % (rt_filePath))
			params.rt_optionLayoutDict = parseLayoutFile(rt_filePath)
			exec("params." + kShowMetaTag + "WidthOnly = 1")
			exec("params." + kShowMetaTag + "HintDir = 1")
			exec("params." + kShowMetaTag + "RowFont = 1")
			exec("params." + kShowMetaTag + "Hints = 0")
			exec("params." + kShowMetaTag + "SideBearings = 0")
			exec("params." + kShowMetaTag + "V_SideBearings = 0")

		elif arg == "-select_hints":
			i = i +1
			indexList = sys.argv[i]
			indexList = re.findall(r"([^,]+)", indexList)
			try:
				indexList = map(int, indexList)
				params.rt_hintTableList = indexList
			except ValueError:
				raise OptionParseError("Option Error: in \" -select_hints %s\,  one of the indices in the argument list is not an integer." %   sys.argv[i])

		elif arg == "-baseline":
			i = i +1
			try:
				params.userBaseLine= int( sys.argv[i])
			except ValueError:
				raise OptionParseError("Option Error: -pt must be followed by an integer number")
		elif arg == "-v":
			exec("params." + kShowMetaTag + "SideBearings = 0")
			exec("params." + kShowMetaTag + "V_SideBearings = 1")
		elif arg == "-vh":
			exec("params." + kShowMetaTag + "SideBearings = 1")
			exec("params." + kShowMetaTag + "V_SideBearings = 1")
		elif arg[0] == "-":
			raise OptionParseError("Option Error: Unknown option <%s>." %  arg)
		else:
			params.rt_fileList += [arg]
		i  += 1
	if not params.rt_fileList:
		raise OptionParseError("Option Error: You must provide at least one font file path.")

	return params


def getGlyphID(glyphTag, fontGlyphList):
	glyphID = None
	try:
		glyphID = int(glyphTag)
		glyphName = fontGlyphList[glyphID]
	except IndexError:
		pass
	except ValueError:
		try:
			glyphID = fontGlyphList.index(glyphTag)
		except IndexError:
			pass
		except ValueError:
			pass
	return glyphID

def getGlyphNames(glyphTag, fontGlyphList, fontFileName):
	glyphNameList = []
	rangeList = glyphTag.split("-")
	prevGID = getGlyphID(rangeList[0], fontGlyphList)
	if prevGID is None:
		if len(rangeList) > 1:
			logMsg( "\tWarning: glyph ID <%s> in range %s from glyph selection list option is not in font. <%s>." % (rangeList[0], glyphTag, fontFileName))
		else:
			logMsg( "\tWarning: glyph ID <%s> from glyph selection list option is not in font. <%s>." % (rangeList[0], fontFileName))
		return None
	glyphNameList.append(fontGlyphList[prevGID])

	for glyphTag2 in rangeList[1:]:
		#import pdb
		#pdb.set_trace()
		gid = getGlyphID(glyphTag2, fontGlyphList)
		if gid is None:
			logMsg( "\tWarning: glyph ID <%s> in range %s from glyph selection list option is not in font. <%s>." % (glyphTag2, glyphTag, fontFileName))
			return None
		for i in range(prevGID+1, gid+1):
			glyphNameList.append(fontGlyphList[i])
		prevGID = gid

	return glyphNameList

def filterGlyphList(params, fontGlyphList, fontFileName):
	# Return the list of glyphs which are in the intersection of the argument list and the glyphs in the font
	# Complain about glyphs in the argument list which are not in the font.
	if not params.rt_optionGlyphList:
		glyphList = fontGlyphList
	else:
		# expand ranges:
		glyphList = []
		for glyphTag in params.rt_optionGlyphList:
			glyphNames = getGlyphNames(glyphTag, fontGlyphList, fontFileName)
			if glyphNames != None:
				glyphList.extend(glyphNames)
	return glyphList


def openFile(path, txPath):
	# If input font is  CFF or PS, build a dummy ttFont.
	tempPathCFF = None
	cffPath = None

	# If it is CID-keyed font, we need to convert it to a name-keyed font. This is a hack, but I really don't want to add CID support to
	# the very simple-minded PDF library.
	command="%s   -dump -0  \"%s\" 2>&1" % (txPath, path)
	report = fdkutils.runShellCmd(command)
	if "CIDFontName" in report:
		tfd,tempPath1 = tempfile.mkstemp()
		os.close(tfd)
		command="%s   -t1 -decid -usefd 0  \"%s\" \"%s\" 2>&1" % (txPath, path, tempPath1)
		report = fdkutils.runShellCmd(command)
		if "fatal" in report:
			logMsg(report)
			logMsg("Failed to convert CID-keyed font %s to a temporary Typ1 data file." % path)

		tfd,tempPathCFF = tempfile.mkstemp()
		os.close(tfd)
		command="%s   -cff +b  \"%s\" \"%s\" 2>&1" % (txPath, tempPath1, tempPathCFF)
		report = fdkutils.runShellCmd(command)
		if "fatal" in report:
			logMsg(report)
			logMsg("Failed to convert CID-keyed font %s to a temporary CFF data file." % path)
		cffPath = tempPathCFF
		os.remove(tempPath1)

	elif os.path.isdir(path):
		# See if it is a UFO font by truing to dump it.
		command="%s   -dump -0  \"%s\" 2>&1" % (txPath, path)
		report = fdkutils.runShellCmd(command)
		if not "sup.srcFontType" in report:
			logMsg(report)
			logMsg("Failed to open directory %s as a UFO font." % path)

		tfd,tempPathCFF = tempfile.mkstemp()
		os.close(tfd)
		command="%s   -cff +b  \"%s\" \"%s\" 2>&1" % (txPath, path, tempPathCFF)
		report = fdkutils.runShellCmd(command)
		if "fatal" in report:
			logMsg(report)
			logMsg("Failed to convert ufo font %s to a temporary CFF data file." % path)
		cffPath = tempPathCFF

	else:
		try:
			with open(path, "rb") as ff:
				head = ff.read(10)
		except (IOError, OSError):
			traceback.print_exc()
			raise FontError("Failed to open and read font file %s. Check file/directory permissions." % path)

		if len(head) < 10:
			raise FontError("Error: font file was zero size: may be a resource fork font, which this program does not process. <%s>." % path)

		# it is an OTF/TTF font, can process file directly
		elif head[:4] in (b"OTTO", b"true", b"\0\1\0\0"):
			try:
				ttFont = TTFont(path)
			except (IOError, OSError):
				raise FontError("Error opening or reading from font file <%s>." % path)
			except TTLibError:
				raise

			if 'CFF ' not in ttFont and 'glyf' not in ttFont:
				raise FontError("Error: font is not a CFF or TrueType font <%s>." % path)

			return ttFont, tempPathCFF

		# It is not an SFNT file.
		elif head[0:2] == b'\x01\x00': # CFF file
			cffPath = path

		elif b"%" not in head:
			# not a PS file either
			logMsg("Font file must be a PS, CFF or OTF  fontfile: %s." % path)
			raise FontError("Font file must be PS, CFF or OTF file: %s." % path)

		else:  # It is a PS file. Convert to CFF.
			tfd,tempPathCFF = tempfile.mkstemp()
			os.close(tfd)
			cffPath = tempPathCFF
			command="%s   -cff +b  \"%s\" \"%s\" 2>&1" % (txPath, path, tempPathCFF)
			report = fdkutils.runShellCmd(command)
			if "fatal" in report:
				logMsg("Attempted to convert font %s  from PS to a temporary CFF data file." % path)
				logMsg(report)
				raise FontError("Failed to convert PS font %s to a temp CFF font." % path)

	# now package the CFF font as an OTF font
	with open(cffPath, "rb") as ff:
		data = ff.read()
	try:
		ttFont = TTFont()
		cffModule = getTableModule('CFF ')
		cffTable = cffModule.table_C_F_F_('CFF ')
		ttFont['CFF '] = cffTable
		cffTable.decompile(data, ttFont)
	except:
		traceback.print_exc()
		logMsg("Attempted to read font %s  as CFF." % path)
		raise FontError("Error parsing font file <%s>." % path)
	return ttFont, tempPathCFF


def proofMakePDF(pathList, params, txPath):
	#    use fontTools library to open font and extract CFF table.
	#    If error, skip font and report error.
	if params.rt_doFontSet:
		pdfFontList = []
		logMsg("")
		logMsg( "Collecting font data from:")
		fontCount = 0
		for path in pathList:
			fontFileName = os.path.basename(path)
			params.rt_filePath = os.path.abspath(path)
			logMsg( "\t%s." % (path))
			try:
				ttFont, tempPathCFF = openFile(path, txPath)
			except FontError:
				print(traceback.format_exception_only(sys.exc_type, sys.exc_value)[-1])
				return
			try:
				fontGlyphList = ttFont.getGlyphOrder()
			except FontError:
				print(traceback.format_exception_only(sys.exc_type, sys.exc_value)[-1])
				return


			#   filter specified list, if any, with font list.
			glyphList = filterGlyphList(params, fontGlyphList, fontFileName)
			if not glyphList:
				raise FontError("Error: selected glyph list is empty for font <%s>." % fontFileName)
			params.rt_reporter = logMsg

			if ttFont.has_key("CFF "):
				pdfFont = otfpdf.txPDFFont(ttFont, params)
			elif ttFont.has_key("glyf"):
				pdfFont = ttfpdf.txPDFFont(ttFont, params)
			else:
				logMsg( "Quitting. Font type is not recognized. %s.." % (path))
				break
			if tempPathCFF:
				pdfFont.path = tempPathCFF
			pdfFontList.append([glyphList, pdfFont, tempPathCFF])

		doProgressBar = not params.quietMode
		pdfFilePath = makeFontSetPDF(pdfFontList, params, doProgressBar)
		for entry in pdfFontList:
			tempPathCFF =  entry[2]
			pdfFont = entry[1]
			pdfFont.clientFont.close()
			if tempPathCFF:
				os.remove(tempPathCFF)

		logMsg( "Wrote proof file %s. End time: %s." % (pdfFilePath, time.asctime()))
		if pdfFilePath and params.openPDFWhenDone:
			if curSystem == "Windows":
				command = 'start "" "%s"' % pdfFilePath
				fdkutils.runShellCmdLogging(command)
			elif curSystem == "Linux":
				command = 'xdg-open "%s" &' % pdfFilePath
				fdkutils.runShellCmdLogging(command)
			else:
				command = 'open "%s" &' % pdfFilePath
				fdkutils.runShellCmdLogging(command)

	else:
		tmpList = []
		for path in pathList:
			fontFileName = os.path.basename(path)
			params.rt_filePath = os.path.abspath(path)
			if not params.quietMode:
				logMsg("")
				logMsg( "Proofing font %s. Start time: %s." % (path, time.asctime()))
			try:
				ttFont, tempPathCFF = openFile(path, txPath)
				fontGlyphList = ttFont.getGlyphOrder()
			except FontError:
				print(traceback.format_exception_only(sys.exc_type, sys.exc_value)[-1])
				return

			#   filter specified list, if any, with font list.
			params.rt_glyphList = filterGlyphList(params, fontGlyphList, fontFileName)
			if not params.rt_glyphList:
				raise FontError("Error: selected glyph list is empty for font <%s>." % fontFileName)
			params.rt_reporter = logMsg

			if "CFF " in ttFont:
				pdfFont = otfpdf.txPDFFont(ttFont, params)
			elif "glyf" in ttFont:
				pdfFont = ttfpdf.txPDFFont(ttFont, params)
			else:
				logMsg( "Quitting. Font type is not recognized. %s.." % (path))
				return

			if tempPathCFF:
				pdfFont.path = tempPathCFF
			doProgressBar = not params.quietMode

			pdfFilePath = makePDF(pdfFont, params, doProgressBar)
			ttFont.close()

			if tempPathCFF:
				os.remove(tempPathCFF)

			if not params.quietMode:
				logMsg("Wrote proof file %s. End time: %s." % (pdfFilePath, time.asctime()))
			if pdfFilePath and params.openPDFWhenDone:
				if curSystem == "Windows":
					command = 'start "" "%s"' % pdfFilePath
					fdkutils.runShellCmdLogging(command)
				elif curSystem == "Linux":
					command = 'xdg-open "%s" &' % pdfFilePath
					fdkutils.runShellCmdLogging(command)
				else:
					command = 'open "%s" &' % pdfFilePath
					fdkutils.runShellCmdLogging(command)


def charplot():
	sys.argv.insert(1, "-charplot")
	main()


def digiplot():
	sys.argv.insert(1, "-digiplot")
	main()


def fontplot():
	sys.argv.insert(1, "-fontplot")
	main()


def fontplot2():
	sys.argv.insert(1, "-fontplot2")
	main()


def fontsetplot():
	sys.argv.insert(1, "-fontsetplot")
	main()


def hintplot():
	sys.argv.insert(1, "-hintplot")
	main()


def waterfallplot():
	sys.argv.insert(1, "-waterfallplot")
	main()


def main():
	try:
		txPath = CheckEnvironment()
	except FDKEnvironmentError as e:
		print(e)
		return
	try:
		params = getOptions(FontPDFParams())
	except OptionParseError as e:
		print(e)
		print(__usage__)
		return


	# verify that all files exist.
	haveFiles = 1
	for path in params.rt_fileList:
		if not os.path.exists(path):
			print("File does not exist: <%s>." % path)
			haveFiles = 0
	if not haveFiles:
		return

	try:
		proofMakePDF(params.rt_fileList, params, txPath)
	except (FontError, NotImplementedError) as e:
		print("\t exception %s" % e)
		traceback.print_exc()
	return


if __name__ == '__main__':
	main()
